#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Interfaces"""

import logging
from abc import ABC, abstractmethod
from typing import Protocol

import numpy.typing as npt

from aragog.utilities import FloatOrArray

logger: logging.Logger = logging.getLogger(__name__)


class PropertyProtocol(Protocol):
    """Property protocol"""

    def eval(self, *args) -> FloatOrArray: ...

    def __call__(self, temperature: npt.NDArray, pressure: npt.NDArray) -> FloatOrArray: ...


class PhaseEvaluatorProtocol(Protocol):
    """Phase evaluator protocol"""

    def set_temperature(self, temperature: npt.NDArray) -> None: ...

    def set_pressure(self, pressure: npt.NDArray) -> None: ...

    def update(self) -> None: ...

    def density(self) -> FloatOrArray: ...

    def dTdPs(self) -> npt.NDArray: ...

    def dTdrs(self) -> npt.NDArray: ...

    def gravitational_acceleration(self) -> FloatOrArray: ...

    def heat_capacity(self) -> FloatOrArray: ...

    def kinematic_viscosity(self) -> FloatOrArray: ...

    def melt_fraction(self) -> FloatOrArray: ...

    def latent_heat(self) -> FloatOrArray: ...

    def thermal_conductivity(self) -> FloatOrArray: ...

    def thermal_expansivity(self) -> FloatOrArray: ...

    def viscosity(self) -> FloatOrArray: ...

    def relative_velocity(self) -> FloatOrArray: ...

    def delta_specific_volume(self) -> FloatOrArray: ...

class MixedPhaseEvaluatorProtocol(PhaseEvaluatorProtocol, Protocol):
    """Mixed phase evaluator protocol"""

    def liquidus(self) -> npt.NDArray: ...

    def liquidus_gradient(self) -> npt.NDArray: ...

    def solidus(self) -> npt.NDArray: ...

    def solidus_gradient(self) -> npt.NDArray: ...


class PhaseEvaluatorABC(ABC):
    """Phase evaluator ABC"""

    temperature: npt.NDArray
    pressure: npt.NDArray

    def set_temperature(self, temperature: npt.NDArray) -> None:
        """Sets the temperature."""
        logger.debug("set_temperature = %s", temperature)
        self.temperature = temperature

    def set_pressure(self, pressure: npt.NDArray) -> None:
        """Sets the pressure."""
        logger.debug("set_pressure = %s", pressure)
        self.pressure = pressure

    def update(self) -> None:
        """Updates quantities to avoid repeat, possibly expensive, calculations."""

    @abstractmethod
    def density(self) -> FloatOrArray: ...

    def dTdPs(self) -> npt.NDArray:
        """TODO: Update reference to sphinx: Solomatov (2007), Treatise on Geophysics, Eq. 3.2"""
        dTdPs: npt.NDArray = (
            self.thermal_expansivity() * self.temperature / (self.density() * self.heat_capacity())
        )

        return dTdPs

    def dTdrs(self) -> npt.NDArray:
        dTdrs: npt.NDArray = (
            -self.gravitational_acceleration()
            * self.thermal_expansivity()
            * self.temperature
            / self.heat_capacity()
        )

        return dTdrs

    @abstractmethod
    def gravitational_acceleration(self) -> FloatOrArray: ...

    @abstractmethod
    def heat_capacity(self) -> FloatOrArray: ...

    def kinematic_viscosity(self) -> FloatOrArray:
        viscosity: FloatOrArray = self.viscosity() / self.density()

        return viscosity

    @abstractmethod
    def melt_fraction(self) -> FloatOrArray: ...

    @abstractmethod
    def latent_heat(self) -> FloatOrArray: ...

    @abstractmethod
    def thermal_conductivity(self) -> FloatOrArray: ...

    @abstractmethod
    def thermal_expansivity(self) -> FloatOrArray: ...

    @abstractmethod
    def viscosity(self) -> FloatOrArray: ...

    @abstractmethod
    def relative_velocity(self) -> FloatOrArray: ...

    @abstractmethod
    def delta_specific_volume(self) -> FloatOrArray: ...


#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Parses the configuration file and scales and stores the parameters."""

from __future__ import annotations

import logging
import sys
from dataclasses import Field, dataclass, field, fields
from typing import Any

import numpy as np
import numpy.typing as npt
from scipy import constants
from thermochem import codata
from typed_configparser import ConfigParser

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

logger: logging.Logger = logging.getLogger(__name__)


def _get_dataclass_from_section_name() -> dict[str, Any]:
    """Maps the section names in the configuration data to the dataclasses that stores the data."""
    mapping: dict[str, Any] = {
        "scalings": _ScalingsParameters,
        "solver": _SolverParameters,
        "boundary_conditions": _BoundaryConditionsParameters,
        "mesh": _MeshParameters,
        "energy": _EnergyParameters,
        "initial_condition": _InitialConditionParameters,
        "phase_liquid": _PhaseParameters,
        "phase_solid": _PhaseParameters,
        "phase_mixed": _PhaseMixedParameters,
        # radionuclides are dealt with separately
    }

    return mapping


@dataclass
class _ScalingsParameters:
    """Stores parameters in the scalings section in the configuration data. All units are SI.

    Args:
        radius: Radius in metres. Defaults to 1.
        temperature: Temperature in Kelvin. Defaults to 1.
        density: Density in kg/m^3. Defaults to 1.
        time: Time in seconds. Defaults to 1.

    Attributes:
        radius, m
        temperature, K
        density, kg/m^3
        time, s
        area, m^2
        kinetic_energy_per_volume, J/m^3
        gravitational_acceleration, m/s^2
        heat_capacity, J/kg/K
        heat_flux, W/m^2
        latent_heat_per_mass, J/kg
        power_per_mass, W/kg
        power_per_volume, W/m^3
        pressure, Pa
        temperature_gradient, K/m
        thermal_expansivity, 1/K
        thermal_conductivity, W/m/K
        velocity, m/s
        viscosity, Pa s
        time_years, years
        stefan_boltzmann_constant (non-dimensional)
    """

    radius: float = 1
    temperature: float = 1
    density: float = 1
    time: float = 1
    area: float = field(init=False)
    gravitational_acceleration: float = field(init=False)
    temperature_gradient: float = field(init=False)
    thermal_expansivity: float = field(init=False)
    pressure: float = field(init=False)
    velocity: float = field(init=False)
    kinetic_energy_per_volume: float = field(init=False)
    heat_capacity: float = field(init=False)
    latent_heat_per_mass: float = field(init=False)
    power_per_volume: float = field(init=False)
    power_per_mass: float = field(init=False)
    heat_flux: float = field(init=False)
    thermal_conductivity: float = field(init=False)
    viscosity: float = field(init=False)
    time_years: float = field(init=False)
    stefan_boltzmann_constant: float = field(init=False)

    def __post_init__(self) -> None:
        self.area = np.square(self.radius)
        self.gravitational_acceleration = self.radius / np.square(self.time)
        self.temperature_gradient = self.temperature / self.radius
        self.thermal_expansivity = 1 / self.temperature
        self.pressure = self.density * self.gravitational_acceleration * self.radius
        self.velocity = self.radius / self.time
        self.kinetic_energy_per_volume = self.density * np.square(self.velocity)
        self.heat_capacity = self.kinetic_energy_per_volume / self.density / self.temperature
        self.latent_heat_per_mass = self.heat_capacity * self.temperature
        self.power_per_volume = self.kinetic_energy_per_volume / self.time
        self.power_per_mass = self.power_per_volume / self.density
        self.heat_flux = self.power_per_volume * self.radius
        self.thermal_conductivity = self.power_per_volume * self.area / self.temperature
        self.viscosity = self.pressure * self.time
        self.time_years = self.time / constants.Julian_year  # Equivalent to TIMEYRS C code
        # Stefan-Boltzmann units for dimensional are W/m^2/K^4
        self.stefan_boltzmann_constant: float = codata.value("Stefan-Boltzmann constant")
        self.stefan_boltzmann_constant /= (
            self.power_per_volume * self.radius / np.power(self.temperature, 4)
        )
        logger.debug("scalings = %s", self)


@dataclass
class _BoundaryConditionsParameters:
    """Stores parameters in the boundary_conditions section in the configuration data."""

    outer_boundary_condition: int
    outer_boundary_value: float
    inner_boundary_condition: int
    inner_boundary_value: float
    emissivity: float
    equilibrium_temperature: float
    core_heat_capacity: float
    scalings_: _ScalingsParameters = field(init=False)

    def scale_attributes(self, scalings: _ScalingsParameters) -> None:
        """Scales the attributes.

        Args:
            scalings: scalings
        """
        self.scalings_ = scalings
        self.equilibrium_temperature /= self.scalings_.temperature
        self.core_heat_capacity /= self.scalings_.heat_capacity
        self._scale_inner_boundary_condition()
        self._scale_outer_boundary_condition()

    def _scale_inner_boundary_condition(self) -> None:
        """Scales the inner boundary value.

        Equivalent to CORE_BC in C code.
            1: Simple core cooling
            2: Prescribed heat flux
            3: Prescribed temperature
        """
        if self.inner_boundary_condition == 1:
            self.inner_boundary_value = 0
        elif self.inner_boundary_condition == 2:
            self.inner_boundary_value /= self.scalings_.heat_flux
        elif self.inner_boundary_condition == 3:
            self.inner_boundary_value /= self.scalings_.temperature
        else:
            msg: str = f"inner_boundary_condition = {self.inner_boundary_condition} is unknown"
            raise ValueError(msg)

    def _scale_outer_boundary_condition(self) -> None:
        """Scales the outer boundary value.

        Equivalent to SURFACE_BC in C code.
            1: Grey-body atmosphere
            2: Zahnle steam atmosphere
            3: Couple to atmodeller
            4: Prescribed heat flux
            5: Prescribed temperature
        """
        if self.outer_boundary_condition == 1:
            pass
        elif self.outer_boundary_condition == 2:
            pass
        elif self.outer_boundary_condition == 3:
            pass
        elif self.outer_boundary_condition == 4:
            self.outer_boundary_value /= self.scalings_.heat_flux
        elif self.outer_boundary_condition == 5:
            self.outer_boundary_value /= self.scalings_.temperature
        else:
            msg: str = f"outer_boundary_condition = {self.outer_boundary_condition} is unknown"
            raise ValueError(msg)


@dataclass
class _EnergyParameters:
    """Stores parameters in the energy section"""

    conduction: bool
    convection: bool
    gravitational_separation: bool
    mixing: bool
    radionuclides: bool
    dilatation: bool
    tidal: bool

    tidal_array: npt.NDArray = field(default_factory=lambda:np.array([0.0], dtype=float))

    def scale_attributes(self, scalings: _ScalingsParameters) -> None:
        """Scales the attributes.

        Args:
            scalings: scalings
        """
        self.scalings_ = scalings
        self.tidal_array /= self.scalings_.power_per_mass


@dataclass
class _InitialConditionParameters:
    """Stores the settings in the initial_condition section in the configuration data."""

    initial_condition: int = 1
    surface_temperature: float = 4000
    basal_temperature: float = 4000
    init_file: str = ""
    scalings_: _ScalingsParameters = field(init=False)

    def scale_attributes(self, scalings: _ScalingsParameters) -> None:
        """Scales the attributes.

        Initial condition method
            1: Linear profile
            2: User-defined temperature field (from file)
            3: Adiabatic profile

        Args:
            scalings: scalings
        """
        self.scalings_ = scalings
        self.surface_temperature /= self.scalings_.temperature
        self.basal_temperature /= self.scalings_.temperature

        if self.initial_condition == 2:
            if self.init_file == "":
                msg: str = (f"you must provide an initial temperature file")
                raise ValueError(msg)
            self.init_temperature = np.loadtxt(self.init_file)
            self.init_temperature /= self.scalings_.temperature


@dataclass
class _MeshParameters:
    """Stores parameters in the mesh section in the configuration data."""

    outer_radius: float
    inner_radius: float
    number_of_nodes: int
    mixing_length_profile: str
    core_density: float
    # Static pressure profile is derived from the Adams-Williamson equation of state.
    eos_method: int = 1 # 1: Adams-Williamson / 2: User defined
    surface_density: float = 4000
    gravitational_acceleration: float = 9.81
    adiabatic_bulk_modulus: float = 260e9
    mass_coordinates: bool = False
    eos_file: str = ""
    scalings_: _ScalingsParameters = field(init=False)

    def scale_attributes(self, scalings: _ScalingsParameters) -> None:
        """Scales the attributes

        Args:
            scalings: scalings
        """
        self.scalings_ = scalings
        self.outer_radius /= self.scalings_.radius
        self.inner_radius /= self.scalings_.radius
        self.core_density /= self.scalings_.density
        self.surface_density /= self.scalings_.density
        self.gravitational_acceleration /= self.scalings_.gravitational_acceleration
        self.adiabatic_bulk_modulus /= self.scalings_.pressure

        if self.eos_method == 2:
            if self.eos_file == "":
                msg: str = (f"you must provide a file for setting up equation of state")
                raise ValueError(msg)
            arr = np.loadtxt(self.eos_file)
            self.eos_radius = arr[:,0] / self.scalings_.radius
            self.eos_pressure = arr[:,1] / self.scalings_.pressure
            self.eos_density = arr[:,2] / self.scalings_.density
            self.eos_gravity = arr[:,3] / self.scalings_.gravitational_acceleration
            # Check that provided eos radius roughly match with Aragog mesh
            if ((self.eos_radius[0] < self.inner_radius) or
                (self.eos_radius[-1] > self.outer_radius) or
                (self.eos_radius[-1]-self.eos_radius[0]) < 0.75*(self.outer_radius-self.inner_radius)):
                msg: str = (f"Radius array in EOS file: Values out of range.")
                raise ValueError(msg)

@dataclass
class _PhaseMixedParameters:
    """Stores settings in the phase_mixed section in the configuration data."""

    latent_heat_of_fusion: float
    rheological_transition_melt_fraction: float
    rheological_transition_width: float
    solidus: str
    liquidus: str
    phase: str
    phase_transition_width: float
    grain_size: float
    scalings_: _ScalingsParameters = field(init=False)

    def scale_attributes(self, scalings: _ScalingsParameters) -> None:
        """Scales the attributes

        Args:
            scalings: scalings
        """
        self.scalings_ = scalings
        self.latent_heat_of_fusion /= self.scalings_.latent_heat_per_mass
        self.grain_size /= self.scalings_.radius


@dataclass
class _PhaseParameters:
    """Stores settings in a phase section in the configuration data.

    This is used to store settings from phase_liquid and phase_solid.
    """

    density: float | str
    heat_capacity: float | str
    melt_fraction: float
    thermal_conductivity: float | str
    thermal_expansivity: float | str
    viscosity: float | str
    scalings_: _ScalingsParameters = field(init=False)

    def scale_attributes(self, scalings: _ScalingsParameters) -> None:
        """Scales the attributes if they are numbers.

        Args:
            scalings: scalings
        """
        self.scalings_ = scalings
        cls_fields: tuple[Field, ...] = fields(self.__class__)
        for field_ in cls_fields:
            value: Any = getattr(self, field_.name)
            try:
                scaling: float = getattr(self.scalings_, field_.name)
                scaled_value = value / scaling
                setattr(self, field_.name, scaled_value)
                logger.info(
                    "%s is a number (value = %s, scaling = %s, scaled_value = %s)",
                    field_.name,
                    value,
                    scaling,
                    scaled_value,
                )
            except AttributeError:
                logger.info("No scaling found for %s", field_.name)
            except TypeError:
                logger.info(
                    "%s is a string (path to a filename) so the data will be scaled later",
                    field_.name,
                )


@dataclass
class _Radionuclide:
    """Stores the settings in a radionuclide section in the configuration data."""

    name: str
    t0_years: float
    abundance: float
    concentration: float
    heat_production: float
    half_life_years: float
    scalings_: _ScalingsParameters = field(init=False)

    def scale_attributes(self, scalings: _ScalingsParameters) -> None:
        """Scales the attributes.

        Args:
            scalings: scalings
        """
        self.scalings_ = scalings
        self.t0_years /= self.scalings_.time_years
        self.concentration *= 1e-6  # to mass fraction
        self.heat_production /= self.scalings_.power_per_mass
        self.half_life_years /= self.scalings_.time_years

    def get_heating(self, time: npt.NDArray | float) -> npt.NDArray | float:
        """Radiogenic heating

        Args:
            time: Time

        Returns:
            Radiogenic heat production (power per unit mass) as a float if time is a float,
                otherwise a numpy row array where each entry in the row is associated
                with a single time in the time array.
        """
        arg: npt.NDArray | float = np.log(2) * (self.t0_years - time) / self.half_life_years
        heating: npt.NDArray | float = (
            self.heat_production * self.abundance * self.concentration * np.exp(arg)
        )

        return heating


@dataclass
class _SolverParameters:
    """Stores settings in the solver section in the configuration data."""

    start_time: float
    end_time: float
    atol: float
    rtol: float
    scalings_: _ScalingsParameters = field(init=False)
    tsurf_poststep_change: float = 30.0
    event_triggering:bool = False

    def scale_attributes(self, scalings: _ScalingsParameters) -> None:
        self.scalings_ = scalings
        self.start_time /= self.scalings_.time_years
        self.end_time /= self.scalings_.time_years
        self.tsurf_poststep_change /= self.scalings_.temperature


@dataclass(kw_only=True)
class Parameters:
    """Assembles all the parameters.

    The parameters in each section are scaled here to ensure that all the parameters are scaled
    (non-dimensionalised) consistently with each other.
    """

    boundary_conditions: _BoundaryConditionsParameters
    energy: _EnergyParameters
    initial_condition: _InitialConditionParameters
    mesh: _MeshParameters
    phase_solid: _PhaseParameters
    phase_liquid: _PhaseParameters
    phase_mixed: _PhaseMixedParameters
    radionuclides: list[_Radionuclide]
    scalings: _ScalingsParameters
    solver: _SolverParameters

    def __post_init__(self):
        cls_fields: tuple[Field, ...] = fields(self.__class__)
        for field_ in cls_fields:
            data = getattr(self, field_.name)
            # Dataclass
            if hasattr(data, "scale_attributes"):
                data.scale_attributes(self.scalings)
            # List of dataclasses
            elif isinstance(data, list):
                for entry in data:
                    if hasattr(entry, "scale_attributes"):
                        entry.scale_attributes(self.scalings)

    @classmethod
    def from_file(cls, *filenames) -> Self:
        """Parses the parameters in a configuration file(s)

        Args:
            *filenames: Filenames of the configuration data
        """
        parser: ConfigParser = ConfigParser()
        parser.read(*filenames)

        init_dict: dict[str, Any] = {}
        for section_name, dataclass_ in _get_dataclass_from_section_name().items():
            init_dict[section_name] = parser.parse_section(
                using_dataclass=dataclass_, section_name=section_name
            )
        radionuclides: list[_Radionuclide] = []
        for radionuclide_section in cls.radionuclide_sections(parser):
            radionuclide = parser.parse_section(
                using_dataclass=_Radionuclide, section_name=radionuclide_section
            )
            radionuclides.append(radionuclide)

        init_dict["radionuclides"] = radionuclides

        return cls(**init_dict)  # Unpacking gives required arguments so pylint: disable=E1125

    @staticmethod
    def radionuclide_sections(parser: ConfigParser) -> list[str]:
        """Section names relating to radionuclides

        Sections relating to radionuclides must have the prefix radionuclide_
        """
        return [
            parser[section].name
            for section in parser.sections()
            if section.startswith("radionuclide_")
        ]



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Output"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt
import netCDF4 as nc
import numpy as np
import numpy.typing as npt
from scipy.optimize import OptimizeResult

from aragog import __version__
from aragog.parser import Parameters
from aragog.solver import Evaluator
from aragog.utilities import FloatOrArray

logger: logging.Logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from aragog.solver import Solver, State


class Output:
    """Stores inputs and outputs of the models."""

    def __init__(self, solver: Solver):
        self.solver: Solver = solver
        self.parameters: Parameters = solver.parameters
        self.solution: OptimizeResult = self.solver.solution
        self.evaluator: Evaluator = self.solver.evaluator
        self.state: State = self.solver.state

    @property
    def shape_basic(self) -> npt.NDArray:
        """Shape of the basic data"""
        return np.array([self.solution.y.shape[0] + 1, self.solution.y.shape[1]])

    @property
    def shape_staggered(self) -> npt.NDArray:
        """Shape of the staggered data"""
        return self.solution.y.shape

    @property
    def conductive_heat_flux_basic(self) -> npt.NDArray:
        """Conductive heat flux"""
        return self.state.conductive_heat_flux() * self.parameters.scalings.heat_flux

    @property
    def convective_heat_flux_basic(self) -> npt.NDArray:
        """Convective heat flux"""
        return self.state.convective_heat_flux() * self.parameters.scalings.heat_flux

    @property
    def gravitational_separation_heat_flux_basic(self) -> npt.NDArray:
        """Gravitational separation heat flux"""
        return (
            self.state.gravitational_separation_mass_flux()
            * self.state.phase_basic.latent_heat()
            * self.parameters.scalings.heat_flux
        )

    @property
    def mixing_heat_flux_basic(self) -> npt.NDArray:
        """Convective mixing heat flux"""
        return (
            self.state.mixing_mass_flux()
            * self.state.phase_basic.latent_heat()
            * self.parameters.scalings.heat_flux
        )

    @property
    def total_heat_flux_basic(self) -> npt.NDArray:
        """Conductive heat flux"""
        return self.state.heat_flux() * self.parameters.scalings.heat_flux

    @property
    def density_basic(self) -> npt.NDArray:
        """Density"""
        return (
            self.state.phase_basic.density()
            * np.ones(self.shape_basic)
            * self.parameters.scalings.density
        )

    @property
    def dTdr(self) -> npt.NDArray:
        """dTdr"""
        return self.solver.state.dTdr() * self.parameters.scalings.temperature_gradient

    @property
    def dTdrs(self) -> npt.NDArray:
        """dTdrs"""
        return (  # FIXME
            self.state.phase_basic.dTdrs() * self.parameters.scalings.temperature_gradient
        )

    @property
    def heat_capacity_basic(self) -> npt.NDArray:
        """Heat capacity"""
        return (
            self.state.phase_basic.heat_capacity()
            * np.ones(self.shape_basic)
            * self.parameters.scalings.heat_capacity
        )

    @property
    def heating(self) -> npt.NDArray:
        """Internal heat generation at staggered nodes"""
        return self.state.heating * self.parameters.scalings.power_per_mass

    @property
    def heating_radio(self) -> npt.NDArray:
        """Internal heat generation from radioactive decay at staggered nodes"""
        return self.state.heating_radio * self.parameters.scalings.power_per_mass

    @property
    def heating_dilatation(self) -> npt.NDArray:
        """Internal heat generation from dilatation/compression at staggered nodes"""
        return self.state.heating_dilatation * self.parameters.scalings.power_per_mass

    @property
    def heating_tidal(self) -> npt.NDArray:
        """Internal heat generation from tidal heat dissipation at staggered nodes"""
        return self.state.heating_tidal * self.parameters.scalings.power_per_mass

    @property
    def liquidus_K_staggered(self) -> npt.NDArray:
        """Liquidus"""
        return self.evaluator.phases.mixed.liquidus() * self.parameters.scalings.temperature

    @property
    def melt_fraction_staggered(self) -> FloatOrArray:
        """Melt fraction on the staggered mesh"""
        return self.state.phase_staggered.melt_fraction()

    @property
    def melt_fraction_basic(self) -> FloatOrArray:
        """Melt fraction on the basic mesh"""
        return self.state.phase_basic.melt_fraction()

    @property
    def rheological_front(self) -> float:
        """Rheological front at the last solve iteration given user defined threshold.
        It is defined as a dimensionless distance with respect to the outer radius.
        """

        _phi_global = float(self.melt_fraction_global)

        # if global melt fraction is close to one everywhere (magma ocean) rf is the inner radius
        if _phi_global > 0.99:
            rf: float = self.evaluator.mesh.basic.radii[0]
        # if global melt fraction is close to zero everywhere (solidified) rf is the outer radius
        elif _phi_global < 0.01:
            rf = self.evaluator.mesh.basic.radii[-1]
        # general case
        else:
            idx = np.argmin(
                np.abs(
                    self.melt_fraction_basic[:, -1]
                    - self.parameters.phase_mixed.rheological_transition_melt_fraction
                )
            )
            rf = self.evaluator.mesh.basic.radii[idx]

        # Return dimensionless rheological front
        return (self.evaluator.mesh.basic.radii[-1] - rf) / self.evaluator.mesh.basic.radii[-1]

    @property
    def melt_fraction_global(self) -> float:
        """Volume-averaged melt fraction"""
        phase_to_use = self.evaluator._parameters.phase_mixed.phase
        if phase_to_use == "mixed" or phase_to_use == "composite":
            out = self.evaluator.mesh.volume_average(self.melt_fraction_staggered[:, -1])
        else:
            out = self.melt_fraction_staggered
        return out

    @property
    def radii_km_basic(self) -> npt.NDArray:
        """Radii of the basic mesh in km"""
        return self.evaluator.mesh.basic.radii * self.parameters.scalings.radius * 1.0e-3

    @property
    def radii_km_staggered(self) -> npt.NDArray:
        """Radii of the staggered mesh in km"""
        return self.evaluator.mesh.staggered.radii * self.parameters.scalings.radius * 1.0e-3

    @property
    def mass_radii_km_basic(self) -> npt.NDArray:
        """Mass radii of the basic mesh in km"""
        return self.evaluator.mesh.basic.mass_radii * self.parameters.scalings.radius * 1.0e-3

    @property
    def mass_radii_km_staggered(self) -> npt.NDArray:
        """Mass radii of the staggered mesh in km"""
        return self.evaluator.mesh.staggered.mass_radii * self.parameters.scalings.radius * 1.0e-3

    @property
    def pressure_GPa_basic(self) -> npt.NDArray:
        """Pressure of the basic mesh in GPa"""
        return self.evaluator.mesh.basic_pressure * self.parameters.scalings.pressure * 1.0e-9

    @property
    def pressure_GPa_staggered(self) -> npt.NDArray:
        """Pressure of the staggered mesh in GPa"""
        return (
            self.evaluator.mesh.staggered_pressure * self.parameters.scalings.pressure * 1.0e-9
        )

    @property
    def mantle_mass(self) -> float:
        """Mantle mass computed from the AdamsWilliamsonEOS"""
        return np.sum(self.mass_staggered)

    @property
    def mass_staggered(self) -> npt.NDArray:
        """Mass of each layer on staggered mesh"""
        return (
            # shells centred on staggered nodes
            self.evaluator.mesh.staggered_effective_density
            * self.evaluator.mesh.basic.volume
            * self.parameters.scalings.density
            * np.power(self.parameters.scalings.radius, 3)
        )

    @property
    def core_mass(self) -> float:
        """Core mass computed with constant density"""

        # core radius
        R_core = self.evaluator.mesh.basic.inner_boundary * self.parameters.scalings.radius

        # core volume
        volume = 4 * np.pi * (R_core**3) / 3

        # core density
        rho = self.parameters.scalings.density * self.parameters.mesh.core_density

        # core mass
        return rho * volume

    @property
    def solidus_K_staggered(self) -> npt.NDArray:
        """Solidus"""
        return self.evaluator.phases.mixed.solidus() * self.parameters.scalings.temperature

    @property
    def super_adiabatic_temperature_gradient_basic(self) -> npt.NDArray:
        """Super adiabatic temperature gradient"""
        return (
            self.state.super_adiabatic_temperature_gradient
            * self.parameters.scalings.temperature_gradient
        )

    @property
    def temperature_K_basic(self) -> npt.NDArray:
        """Temperature of the basic mesh in K"""
        return self.state.temperature_basic * self.parameters.scalings.temperature

    @property
    def temperature_K_staggered(self) -> npt.NDArray:
        """Temperature of the staggered mesh in K"""
        return self.solver.temperature_staggered

    @property
    def thermal_expansivity_basic(self) -> npt.NDArray:
        """Thermal expansivity"""
        return (
            self.state.phase_basic.thermal_expansivity()
            * np.ones(self.shape_basic)
            * self.parameters.scalings.thermal_expansivity
        )

    @property
    def log10_viscosity_basic(self) -> npt.NDArray:
        """Viscosity of the basic mesh"""
        return np.log10(
            self.state.phase_basic.viscosity()
            * self.parameters.scalings.viscosity
            * np.ones(self.shape_basic)
        )

    @property
    def log10_viscosity_staggered(self) -> npt.NDArray:
        """Viscosity of the staggered mesh"""
        return np.log10(
            self.state.phase_staggered.viscosity()
            * self.parameters.scalings.viscosity
            * np.ones(self.shape_staggered)
        )

    @property
    def solution_top_temperature(self) -> float:
        """Solution (last iteration) temperature at the top of the domain (planet surface)"""
        return self.temperature_K_basic[-1, -1]

    @property
    def times(self) -> npt.NDArray:
        """Times in years"""
        return self.solution.t * self.parameters.scalings.time_years

    @property
    def time_range(self) -> float:
        return self.times[-1] - self.times[0]

    def write_at_time(self, file_path: str, tidx: int = -1, compress: bool = False) -> None:
        """Write the state of the model at a particular time to a NetCDF4 file on the disk.

        Args:
            file_path: Path to the output file
            tidx: Index on the time axis at which to access the data
            compress: Whether to compress the data
        """

        logger.debug("Writing i=%d NetCDF file to %s", tidx, file_path)

        # Update the state
        assert self.solution is not None
        self.state.update(self.solution.y, self.solution.t)

        # Open the dataset
        ds: nc.Dataset = nc.Dataset(file_path, mode="w")

        # Metadata
        ds.description = "Aragog output data"
        ds.argog_version = __version__

        # Function to save scalar quantities
        def _add_scalar_variable(key: str, value: float, units: str):
            ds.createVariable(key, np.float64)
            ds[key][0] = float(value)
            ds[key].units = units

        # Save scalar quantities
        _add_scalar_variable("time", self.times[tidx], "yr")
        _add_scalar_variable("phi_global", self.melt_fraction_global, "")
        _add_scalar_variable("mantle_mass", self.mantle_mass, "kg")
        _add_scalar_variable("rheo_front", self.rheological_front, "")

        # Create dimensions for mesh quantities
        ds.createDimension("basic", self.shape_basic[0])
        ds.createDimension("staggered", self.shape_staggered[0])

        # Function to save mesh quantities
        def _add_mesh_variable(key: str, some_property: Any, units: str):
            if key[-2:] == "_b":
                mesh = "basic"
            elif key[-2:] == "_s":
                mesh = "staggered"
            else:
                raise KeyError(f"NetCDF variable name must end in _b or _s: {key}")
            ds.createVariable(
                key,
                np.float64,
                (mesh,),
                compression="zlib" if compress else None,
                shuffle = True,
                complevel = 4
            )
            ds[key][:] = some_property[:, tidx]
            ds[key].units = units

        # Save mesh quantities
        _add_mesh_variable("radius_b", self.radii_km_basic, "km")
        _add_mesh_variable("pres_b", self.pressure_GPa_basic, "GPa")
        _add_mesh_variable("temp_b", self.temperature_K_basic, "K")
        _add_mesh_variable("phi_b", self.melt_fraction_basic, "")
        _add_mesh_variable("Fcond_b", self.conductive_heat_flux_basic, "W m-2")
        _add_mesh_variable("Fconv_b", self.convective_heat_flux_basic, "W m-2")
        _add_mesh_variable("Fgrav_b", self.gravitational_separation_heat_flux_basic, "W m-2")
        _add_mesh_variable("Fmix_b", self.mixing_heat_flux_basic, "W m-2")
        _add_mesh_variable("Ftotal_b", self.total_heat_flux_basic, "W m-2")
        _add_mesh_variable("log10visc_b", self.log10_viscosity_basic, "Pa s")
        _add_mesh_variable("log10visc_s", self.log10_viscosity_staggered, "Pa s")
        _add_mesh_variable("density_b", self.density_basic, "kg m-3")
        _add_mesh_variable("heatcap_b", self.heat_capacity_basic, "J kg-1 K-1")
        _add_mesh_variable("mass_s", self.mass_staggered, "kg")
        _add_mesh_variable("Hradio_s", self.heating_radio, "W kg-1")
        _add_mesh_variable("Hvol_s", self.heating_dilatation, "W kg-1")
        _add_mesh_variable("Htidal_s", self.heating_tidal, "W kg-1")
        _add_mesh_variable("Htotal_s", self.heating, "W kg-1")

        # Close the dataset
        ds.close()

    def plot(self, num_lines: int = 11, figsize: tuple = (25, 10)) -> None:
        """Plots the solution with labelled lines according to time.

        Args:
            num_lines: Number of lines to plot. Defaults to 11.
            figsize: Size of the figure. Defaults to (25, 10).
        """
        assert self.solution is not None

        self.state.update(self.solution.y, self.solution.t)

        _, axs = plt.subplots(1, 10, sharey=True, figsize=figsize)

        # Ensure there are at least 2 lines to plot (first and last).
        num_lines = max(2, num_lines)

        # Calculate the time step based on the total number of lines.
        time_step: float = self.time_range / (num_lines - 1)

        # plot temperature
        try:
            axs[0].scatter(self.liquidus_K_staggered, self.pressure_GPa_staggered)
            axs[0].scatter(self.solidus_K_staggered, self.pressure_GPa_staggered)
        except AttributeError:
            pass

        # Plot the first line.
        def plot_times(ax, x: npt.NDArray, y: npt.NDArray) -> None:
            # If `x` is a float, create an array of the same length as `y` filled with `x`
            if np.isscalar(x):
                x = np.full((len(y), len(self.times)), x, dtype=np.float64)
            elif x.ndim == 1:
                # If `x` is 1D, reshape it or repeat it across the time dimension
                x = np.tile(x[:, np.newaxis], (1, len(self.times)))

            label_first: str = f"{self.times[0]:.2f}"
            ax.plot(x[:, 0], y, label=label_first)

            # Loop through the selected lines and plot each with a label.
            for i in range(1, num_lines - 1):
                desired_time: float = self.times[0] + i * time_step
                # Find the closest available time step.
                closest_time_index: np.intp = np.argmin(np.abs(self.times - desired_time))
                time: float = self.times[closest_time_index]
                label: str = f"{time:.2f}"  # Create a label based on the time.
                ax.plot(
                    x[:, closest_time_index],
                    y,
                    label=label,
                )

            # Plot the last line.
            times_end: float = self.times[-1]
            label_last: str = f"{times_end:.2f}"
            ax.plot(x[:, -1], y, label=label_last)

        plot_times(axs[0], self.temperature_K_basic, self.pressure_GPa_basic)
        axs[0].set_ylabel("Pressure (GPa)")
        axs[0].set_xlabel("Temperature (K)")
        axs[0].set_title("Temperature")

        plot_times(axs[1], self.melt_fraction_staggered, self.pressure_GPa_staggered)
        axs[1].set_xlabel("Melt fraction")
        axs[1].set_title("Melt fraction")

        plot_times(axs[2], self.log10_viscosity_basic, self.pressure_GPa_basic)
        axs[2].set_xlabel("Log10(viscosity)")
        axs[2].set_title("Log10(viscosity)")

        plot_times(axs[3], self.convective_heat_flux_basic, self.pressure_GPa_basic)
        axs[3].set_xlabel("Convective heat flux")
        axs[3].set_title("Convective heat flux")

        plot_times(
            axs[4],
            self.super_adiabatic_temperature_gradient_basic,
            self.pressure_GPa_basic,
        )
        axs[4].set_xlabel("Super adiabatic temperature gradient")
        axs[4].set_title("Super adiabatic temperature gradient")

        plot_times(
            axs[5],
            self.dTdr,
            self.pressure_GPa_basic,
        )
        axs[5].set_xlabel("dTdr")
        axs[5].set_title("dTdr")

        plot_times(
            axs[6],
            self.dTdrs,
            self.pressure_GPa_basic,
        )
        axs[6].set_xlabel("dTdrs")
        axs[6].set_title("dTdrs")

        plot_times(
            axs[7],
            self.density_basic,
            self.pressure_GPa_basic,
        )
        axs[7].set_xlabel("Density")
        axs[7].set_title("Density")

        plot_times(
            axs[8],
            self.heat_capacity_basic,
            self.pressure_GPa_basic,
        )
        axs[8].set_xlabel("Heat capacity")
        axs[8].set_title("Heat capacity")

        plot_times(
            axs[9],
            self.thermal_expansivity_basic,
            self.pressure_GPa_basic,
        )
        axs[9].set_xlabel("Thermal expansivity")
        axs[9].set_title("Thermal expansivity")

        # Shrink current axis by 20% to allow space for the legend.
        # box = ax.get_position()
        # ax.set_position((box.x0, box.y0, box.width * 0.8, box.height))

        # axs[0].grid(True)

        legend = axs[2].legend()  # (loc="center left", bbox_to_anchor=(1, 0.5))
        legend.set_title("Time (yr)")
        plt.gca().invert_yaxis()

        plt.show()



import click


@click.group()
def cli():
    pass


@click.group()
def download():
    """Download data"""
    pass


@click.command()
def all():
    """Download all lookup table data."""
    from .data import DownloadLookupTableData

    DownloadLookupTableData()


@click.command()
def env():
    """Show environment variables and locations"""
    from .data import FWL_DATA_DIR

    click.echo(f"FWL_DATA location: {FWL_DATA_DIR}")


cli.add_command(download)
download.add_command(all)
cli.add_command(env)

if __name__ == "__main__":
    cli()



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Utilities"""

from __future__ import annotations

from cProfile import Profile
from functools import wraps
from pathlib import Path
from pstats import SortKey, Stats
from typing import Any, TypeVar

import numpy as np
import numpy.typing as npt
import pandas as pd

MultiplyT = TypeVar("MultiplyT", float, npt.NDArray, pd.Series, pd.DataFrame)

FloatOrArray = float | npt.NDArray


def profile_decorator(func):
    """Decorator to profile a function"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        with Profile() as profile:
            result = func(*args, **kwargs)
        stats = Stats(profile).strip_dirs().sort_stats(SortKey.TIME)
        stats.print_stats()
        return result

    return wrapper


def is_file(value: Any) -> bool:
    """Checks if value is a file.

    Args:
        value: Object to be checked

    Returns:
        True if the value is a file, otherwise False
    """
    if isinstance(value, (str, Path)):
        return Path(value).is_file()

    return False


def is_monotonic_increasing(some_array: npt.NDArray) -> np.bool_:
    """Returns True if an array is monotonically increasing, otherwise returns False."""
    return np.all(np.diff(some_array) > 0)


def is_number(value: Any) -> bool:
    """Checks if value is a number.

    Args:
        value: Object to be checked

    Returns:
        True if the value is a number, otherwise False
    """
    try:
        float(value)
        return True

    except (TypeError, ValueError):
        return False


def tanh_weight(value: FloatOrArray, threshold: float, width: float) -> npt.NDArray:
    """Computes the tanh weight for viscosity profile and smoothing.

    Args:
        value: Value
        threshold: Threshold value
        width: Width of smoothing

    Returns:
        weight
    """
    arg: FloatOrArray = (value - threshold) / width
    weight: npt.NDArray = 0.5 * (1.0 + np.tanh(arg))

    return weight


def combine_properties(weight: Any, property1: Any, property2: Any) -> Any:
    """Linear weighting of two quantities.

    Args:
        weight: The weight to apply to property1
        property1: The value of the first property
        property2: The value of the second property

    Returns:
        The combined (weighted) property
    """
    out = weight * property1 + (1.0 - weight) * property2

    return out



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Core classes and functions"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
from scipy.integrate import solve_ivp

from aragog.mesh import Mesh
from aragog.phase import PhaseEvaluatorCollection
from aragog.parser import (
    Parameters,
    _BoundaryConditionsParameters,
    _InitialConditionParameters,
)

if TYPE_CHECKING:
    from aragog.solver import State

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class BoundaryConditions:
    """Boundary conditions

    Args:
        parameters: Parameters
        mesh: Mesh
    """

    _parameters: Parameters
    _mesh: Mesh

    def __post_init__(self):
        self._settings: _BoundaryConditionsParameters = self._parameters.boundary_conditions

    def apply_temperature_boundary_conditions(
        self, temperature: npt.NDArray, temperature_basic: npt.NDArray, dTdr: npt.NDArray
    ) -> None:
        """Conforms the temperature and dTdr at the basic nodes to temperature boundary conditions.

        Args:
            temperature: Temperature at the staggered nodes
            temperature_basic: Temperature at the basic nodes
            dTdr: Temperature gradient at the basic nodes
        """
        # Core-mantle boundary
        if self._settings.inner_boundary_condition == 3:
            temperature_basic[0, :] = self._settings.inner_boundary_value
            dTdr[0, :] = (
                2 * (temperature[0, :] - temperature_basic[0, :])
                / self._mesh.basic.delta_mesh[0]
                * self._mesh.dxidr[0]
            )
        # Surface
        if self._settings.outer_boundary_condition == 5:
            temperature_basic[-1, :] = self._settings.outer_boundary_value
            dTdr[-1, :] = (
                2 * (temperature_basic[-1, :] - temperature[-1, :])
                / self._mesh.basic.delta_mesh[-1]
                * self._mesh.dxidr[-1]
            )

    def apply_temperature_boundary_conditions_melt(
        self, melt_fraction: npt.NDArray, melt_fraction_basic: npt.NDArray, dphidr: npt.NDArray
    ) -> None:
        """Conforms the melt fraction gradient dphidr at the basic nodes
           to temperature boundary conditions.

        Args:
            melt_fraction: Melt fraction at the staggered nodes
            melt_fraction_basic: Melt fraction at the basic nodes
            dphidr: Melt fraction gradient at the basic nodes
        """
        # Core-mantle boundary
        if self._settings.inner_boundary_condition == 3:
            dphidr[0, :] = (
                2 * (melt_fraction[0, :] - melt_fraction_basic[0, :])
                / self._mesh.basic.delta_mesh[0]
                * self._mesh.dxidr[0]
            )
        # Surface
        if self._settings.outer_boundary_condition == 5:
            dphidr[-1, :] = (
                2 * (melt_fraction_basic[-1, :] - melt_fraction[-1, :])
                / self._mesh.basic.delta_mesh[-1]
                * self._mesh.dxidr[-1]
            )

    def apply_flux_boundary_conditions(self, state: State) -> None:
        """Applies the boundary conditions to the state.

        Args:
            state: The state to apply the boundary conditions to
        """
        self.apply_flux_inner_boundary_condition(state)
        self.apply_flux_outer_boundary_condition(state)
        logger.debug("temperature = %s", state.temperature_basic)
        logger.debug("heat_flux = %s", state.heat_flux)

    def apply_flux_outer_boundary_condition(self, state: State) -> None:
        """Applies the flux boundary condition to the state at the outer boundary.

        Args:
            state: The state to apply the boundary conditions to

        Equivalent to SURFACE_BC in C code.
            1: Grey-body atmosphere
            2: Zahnle steam atmosphere
            3: Couple to atmodeller
            4: Prescribed heat flux
            5: Prescribed temperature
        """
        if self._settings.outer_boundary_condition == 1:
            self.grey_body(state)
        elif self._settings.outer_boundary_condition == 2:
            raise NotImplementedError
        elif self._settings.outer_boundary_condition == 3:
            msg: str = "Requires coupling to atmodeller"
            logger.error(msg)
            raise NotImplementedError(msg)
        elif self._settings.outer_boundary_condition == 4:
            state.heat_flux[-1, :] = self._settings.outer_boundary_value
        elif self._settings.outer_boundary_condition == 5:
            pass
        else:
            msg: str = (
                f"outer_boundary_condition = {self._settings.outer_boundary_condition} is unknown"
            )
            raise ValueError(msg)

    def grey_body(self, state: State) -> None:
        """Applies a grey body flux at the surface.

        Args:
            state: The state to apply the boundary conditions to
        """
        state.heat_flux[-1, :] = (
            self._settings.emissivity
            * self._settings.scalings_.stefan_boltzmann_constant
            * (np.power(state.top_temperature, 4) - self._settings.equilibrium_temperature**4)
        )

    def apply_flux_inner_boundary_condition(self, state: State) -> None:
        """Applies the flux boundary condition to the state at the inner boundary.

        Args:
            state: The state to apply the boundary conditions to

        Equivalent to CORE_BC in C code.
            1: Simple core cooling
            2: Prescribed heat flux
            3: Prescribed temperature
        """
        if self._settings.inner_boundary_condition == 1:
            self.core_cooling(state)
        elif self._settings.inner_boundary_condition == 2:
            state.heat_flux[0, :] = self._settings.inner_boundary_value
        elif self._settings.inner_boundary_condition == 3:
            pass
            # raise NotImplementedError
        else:
            msg: str = (
                f"inner_boundary_condition = {self._settings.inner_boundary_condition} is unknown"
            )
            raise ValueError(msg)

    def core_cooling(self, state: State) -> None:
        """Applies a core cooling heat flux according to Eq. (37) of Bower et al., 2018

        Args:
            state: The state to apply the boundary condition to
        """
        core_capacity: float = (
            4
            / 3
            * np.pi
            * np.power(self._mesh.basic.radii[0], 3)
            * self._mesh.settings.core_density
            * self._settings.core_heat_capacity
        )
        cell_capacity = self._mesh.basic.volume[0] * state.capacitance_staggered()[0, :]
        radius_ratio: float = self._mesh.basic.radii[1] / self._mesh.basic.radii[0]
        alpha = np.power(radius_ratio, 2) / ((cell_capacity / (core_capacity * 1.147)) + 1)

        state.heat_flux[0, :] = alpha * state.heat_flux[1, :]


@dataclass
class InitialCondition:
    """Initial condition

    Args:
        parameters: Parameters
        mesh: Mesh
        phases: PhaseEvaluatorCollection
    """

    _parameters: Parameters
    _mesh: Mesh
    _phases: PhaseEvaluatorCollection

    def __post_init__(self):
        self._settings: _InitialConditionParameters = self._parameters.initial_condition

        # Three initialisation methods: linear (1), user-defined field (2) or adiabat (3).
        if self._settings.initial_condition == 1:
            self._temperature: npt.NDArray = self.get_linear()
        elif self._settings.initial_condition == 2:
            if self._mesh.staggered.number_of_nodes == len(self._settings.init_temperature):
                self._temperature = self._settings.init_temperature
            else:
                msg: str = (
                    f"the size of the provided init temperature field does not match \
                    the number of staggered points {self._mesh.staggered.number_of_nodes}"
                )
                raise ValueError(msg)
        elif self._settings.initial_condition == 3:
            self._temperature: npt.NDArray = self.get_adiabat(self._mesh.basic_pressure[:,-1])
        else:
            msg: str = (
                f"initial_condition = {self._settings.initial_condition} is unknown"
            )
            raise ValueError(msg)

        logger.debug("initial staggered temperature = %s", self._temperature)

    @property
    def temperature(self) -> npt.NDArray:
        return self._temperature

    def get_linear(self) -> npt.NDArray:
        """Gets a linear temperature profile

        Returns:
            Linear temperature profile for the staggered nodes
            Only works for uniform spatial mesh.
        """
        temperature_basic: npt.NDArray = np.linspace(
            self._settings.basal_temperature,
            self._settings.surface_temperature,
            self._mesh.basic.number_of_nodes,
        )
        return self._mesh.quantity_at_staggered_nodes(temperature_basic)

    def get_adiabat(self, pressure_basic) -> npt.NDArray:
        """Gets an adiabatic temperature profile by integrating
           the adiatiabatic temperature gradient dTdPs from the surface.
           Uses the set surface temperature.

        Args:
            Pressure field on the basic nodes

        Returns:
            Adiabatic temperature profile for the staggered nodes
        """

        def adiabat_ode(P,T):
            self._phases.active.set_pressure(P)
            self._phases.active.set_temperature(T)
            self._phases.active.update()
            return self._phases.active.dTdPs()

        # flip the pressure field top to bottom
        pressure_basic = np.flip(pressure_basic)

        sol = solve_ivp(
             adiabat_ode, (pressure_basic[0], pressure_basic[-1]),
             [self._settings.surface_temperature], t_eval=pressure_basic,
             method='RK45', rtol=1e-6, atol=1e-9)

        # flip back the temperature field from bottom to top
        temperature_basic = np.flip(sol.y[0])

        # Return temperature field at staggered nodes
        return self._mesh.quantity_at_staggered_nodes(temperature_basic)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Solver"""

from __future__ import annotations

import copy
import logging
import sys
from dataclasses import InitVar, dataclass, field
from pathlib import Path

import numpy as np
import numpy.typing as npt
from scipy.integrate import solve_ivp
from scipy.optimize import OptimizeResult

from aragog.core import BoundaryConditions, InitialCondition
from aragog.interfaces import PhaseEvaluatorProtocol
from aragog.mesh import Mesh
from aragog.parser import Parameters, _EnergyParameters, _Radionuclide
from aragog.phase import PhaseEvaluatorCollection
from aragog.utilities import FloatOrArray

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class State:
    """Stores and updates the state at temperature and pressure.

    Args:
        parameters: Parameters
        evaluator: Evaluator

    Attributes:
        evaluator: Evaluator
        critical_reynolds_number: Critical Reynolds number
        gravitational_separation_flux: Gravitational separation flux at the basic nodes
        heating: Total internal heat production at the staggered nodes (power per unit mass)
        heating_radio: Radiogenic heat production at the staggered nodes (power per unit mass)
        heating_tidal: Tidal heat production at the staggered nodes (power per unit mass)
        heat_flux: Heat flux at the basic nodes (power per unit area)
        inviscid_regime: True if the flow is inviscid and otherwise False, at the basic nodes
        inviscid_velocity: Inviscid velocity at the basic nodes
        is_convective: True if the flow is convecting and otherwise False, at the basic nodes
        reynolds_number: Reynolds number at the basic nodes
        super_adiabatic_temperature_gradient: Super adiabatic temperature gradient at the basic nod
        temperature_basic: Temperature at the basic nodes
        temperature_staggered: Temperature at the staggered nodes
        bottom_temperature: Temperature at the bottom basic node
        top_temperature: Temperature at the top basic node
        viscous_regime: True if the flow is viscous and otherwise False, at the basic nodes
        viscous_velocity: Viscous velocity at the basic nodes
    """

    parameters: InitVar[Parameters]
    _evaluator: Evaluator
    _settings: _EnergyParameters = field(init=False)
    phase_basic: PhaseEvaluatorProtocol = field(init=False)
    phase_staggered: PhaseEvaluatorProtocol = field(init=False)
    _dTdr: npt.NDArray = field(init=False)
    _dphidr: npt.NDArray = field(init=False)
    _eddy_diffusivity: npt.NDArray = field(init=False)
    _mass_flux: npt.NDArray = field(init=False)
    _heat_flux: npt.NDArray = field(init=False)
    _heating: npt.NDArray = field(init=False)
    _heating_radio: npt.NDArray = field(init=False)
    _heating_dilatation: npt.NDArray = field(init=False)
    _heating_tidal: npt.NDArray = field(init=False)
    _is_convective: npt.NDArray = field(init=False)
    _reynolds_number: npt.NDArray = field(init=False)
    _super_adiabatic_temperature_gradient: npt.NDArray = field(init=False)
    _temperature_basic: npt.NDArray = field(init=False)
    _temperature_staggered: npt.NDArray = field(init=False)
    _viscous_velocity: npt.NDArray = field(init=False)
    _inviscid_velocity: npt.NDArray = field(init=False)

    def __post_init__(self, parameters: Parameters):
        self._settings = parameters.energy
        # Sets the pressure since this will not change during a model run. Must deepcopy first
        # because pressure is set as an attribute.
        self.phase_basic = copy.deepcopy(self._evaluator.phases.active)
        self.phase_basic.set_pressure(self._evaluator.mesh.basic_pressure)
        self.phase_staggered = copy.deepcopy(self._evaluator.phases.active)
        self.phase_staggered.set_pressure(self._evaluator.mesh.staggered_pressure)

    def capacitance_staggered(self) -> FloatOrArray:
        capacitance: FloatOrArray = (
            self.phase_staggered.density() * self.phase_staggered.heat_capacity()
        )

        return capacitance

    def conductive_heat_flux(self) -> npt.NDArray:
        r"""Conductive heat flux:

        .. math::
            q_{cond} = -k \frac{\partial T}{\partial r}

        where :math:`k` is thermal conductivity, :math:`T` is temperature, and :math:`r` is radius.
        """
        conductive_heat_flux: npt.NDArray = self.phase_basic.thermal_conductivity() * -self.dTdr()

        return conductive_heat_flux

    def convective_heat_flux(self) -> npt.NDArray:
        r"""Convective heat flux:

        .. math::
            q_{conv} = -\rho c_p \kappa_h \left( \frac{\partial T}{\partial r}
                - \left( \frac{\partial T}{\partial r} \right)_S \right)

        where :math:`\rho` is density, :math:`c_p` is heat capacity at constant pressure,
        :math:`\kappa_h` is eddy diffusivity, :math:`T` is temperature, :math:`r` is radius, and
        :math:`S` is entropy.
        """
        convective_heat_flux: npt.NDArray = (
            self.phase_basic.density()
            * self.phase_basic.heat_capacity()
            * self.eddy_diffusivity()
            * -self._super_adiabatic_temperature_gradient
        )

        return convective_heat_flux

    def gravitational_separation_mass_flux(self) -> npt.NDArray:
        r"""Gravitational separation mass flux:

        .. math::
            j_{grav} = \rho \phi (1 - \phi) v_{rel}

        where :math:`\rho` is density, :math:`\phi` is melt fraction, and
        :math:`v_{rel}` is relative velocity.
        """
        gravitational_separation_mass_flux: npt.NDArray = (
            self.phase_basic.density()
            * self.phase_basic.melt_fraction()
            * (1.0 - self.phase_basic.melt_fraction())
            * self.phase_basic.relative_velocity()
        )
        return gravitational_separation_mass_flux

    def mixing_mass_flux(self) -> npt.NDArray:
        r"""Mixing mass flux:

        .. math::
            j_{cm} = -\rho \kappa_h \frac{\partial \phi}{\partial r}

        where :math:`\rho` is density, :math:`\kappa_h` is eddy diffusivity,
        :math:`\phi` is melt mass fraction, and :math:`r` is radius.
        """
        mixing_mass_flux: npt.NDArray = (
            self.phase_basic.density()
            * self.eddy_diffusivity()
            * -self.dphidr()
        )
        return mixing_mass_flux

    def radiogenic_heating(self, time: float) -> npt.NDArray:
        """Radiogenic heating (constant with radius)

        Args:
            time: Time

        Returns:
            Radiogenic heating (power per unit mass) at each layer of the staggered
                mesh, at a given point in time.
        """

        # Total heat production at a given time (power per unit mass)
        radio_heating_float: float = 0
        for radionuclide in self._evaluator.radionuclides:
            radio_heating_float += radionuclide.get_heating(time)

        # Convert to 1D array (assuming abundances are constant)
        return radio_heating_float * np.ones_like(self.temperature_staggered)

    def dilatation_heating(self) -> npt.NDArray:
        """Dilatation/compression heating (power per unit mass)

        Returns:
            Dilatation/compression heating (power per unit mass) at each layer of the staggered
                mesh, at a given point in time.
        """

        mass_flux_staggered: npt.NDArray = self._evaluator.mesh.quantity_at_staggered_nodes(self._mass_flux)

        dilatation_volume_source: npt.NDArray = (
            self.phase_staggered.gravitational_acceleration()
            * self.phase_staggered.delta_specific_volume()
            * mass_flux_staggered
        )

        return dilatation_volume_source

    def tidal_heating(self) -> npt.NDArray:
        """Tidal heating at each layer of the mantle.

        Args:
            time: Time

        Returns:
            Tidal heating (power per unit mass) at each layer of the staggered
                mesh, at a given point in time.
        """

        length = len(self._settings.tidal_array)

        # Must have correct shape
        if length == 1:
            # scalar
            out = np.ones_like(self.temperature_staggered) * self._settings.tidal_array[0]

        elif length == len(self.temperature_staggered):
            # vector with correct length
            out = np.array([self._settings.tidal_array]).T

        else:
            # vector with invalid length
            raise ValueError(f"Tidal heating array has invalid length {length}")

        return out

    @property
    def critical_reynolds_number(self) -> float:
        """Critical Reynolds number from Abe (1993)"""
        return 9 / 8

    def dTdr(self) -> npt.NDArray:
        return self._dTdr

    def dphidr(self) -> npt.NDArray:
        return self._dphidr

    def eddy_diffusivity(self) -> npt.NDArray:
        return self._eddy_diffusivity

    @property
    def heating(self) -> npt.NDArray:
        """The power generation according to the heat sources specified in the configuration."""
        return self._heating

    @property
    def heating_radio(self) -> npt.NDArray:
        """The radiogenic power generation."""
        return self._heating_radio

    @property
    def heating_dilatation(self) -> npt.NDArray:
        """The heat source through dilation/compression."""
        return self._heating_dilatation

    @property
    def heating_tidal(self) -> npt.NDArray:
        """The tidal power generation."""
        return self._heating_tidal

    @property
    def mass_flux(self) -> npt.NDArray:
        """The total melt mass flux according to the fluxes specified in the configuration."""
        return self._mass_flux

    @property
    def heat_flux(self) -> npt.NDArray:
        """The total heat flux according to the fluxes specified in the configuration."""
        return self._heat_flux

    @property
    def inviscid_regime(self) -> npt.NDArray:
        return self._reynolds_number > self.critical_reynolds_number

    @property
    def inviscid_velocity(self) -> npt.NDArray:
        return self._inviscid_velocity

    @property
    def is_convective(self) -> npt.NDArray:
        return self._is_convective

    @property
    def reynolds_number(self) -> npt.NDArray:
        return self._reynolds_number

    @property
    def super_adiabatic_temperature_gradient(self) -> npt.NDArray:
        return self._super_adiabatic_temperature_gradient

    @property
    def temperature_basic(self) -> npt.NDArray:
        return self._temperature_basic

    @property
    def temperature_staggered(self) -> npt.NDArray:
        return self._temperature_staggered

    @property
    def top_temperature(self) -> npt.NDArray:
        return self._temperature_basic[-1, :]

    @property
    def bottom_temperature(self) -> npt.NDArray:
        return self._temperature_basic[0, :]

    @property
    def viscous_regime(self) -> npt.NDArray:
        return self._reynolds_number <= self.critical_reynolds_number

    @property
    def viscous_velocity(self) -> npt.NDArray:
        return self._viscous_velocity

    def update(self, temperature: npt.NDArray, time: FloatOrArray) -> None:
        """Updates the state.

        The evaluation order matters because we want to minimise the number of evaluations.

        Args:
            temperature: Temperature at the staggered nodes
            pressure: Pressure at the staggered nodes
            time: Time
        """
        logger.debug("Updating the state")

        logger.debug("Setting the temperature profile")
        self._temperature_staggered = temperature
        self._temperature_basic = self._evaluator.mesh.quantity_at_basic_nodes(temperature)
        logger.debug("temperature_basic = %s", self.temperature_basic)
        self._dTdr = self._evaluator.mesh.d_dr_at_basic_nodes(temperature)
        logger.debug("dTdr = %s", self.dTdr())
        self._evaluator.boundary_conditions.apply_temperature_boundary_conditions(
            temperature, self._temperature_basic, self.dTdr()
        )

        logger.debug("Setting the melt fraction profile")
        self.phase_staggered.set_temperature(temperature)
        self.phase_staggered.update()
        self.phase_basic.set_temperature(self._temperature_basic)
        self.phase_basic.update()
        phase_to_use = self._evaluator._parameters.phase_mixed.phase
        if phase_to_use == "mixed" or phase_to_use == "composite":
            self._dphidr = self._evaluator.mesh.d_dr_at_basic_nodes(
                self.phase_staggered.melt_fraction()
            )
            logger.debug("dphidr = %s", self.dphidr())
            self._evaluator.boundary_conditions.apply_temperature_boundary_conditions_melt(
                self.phase_staggered.melt_fraction(), self.phase_basic.melt_fraction(), self._dphidr
            )
        else:
            self._dphidr = np.zeros_like(self._dTdr)

        self._super_adiabatic_temperature_gradient = self.dTdr() - self.phase_basic.dTdrs()
        self._is_convective = self._super_adiabatic_temperature_gradient < 0
        velocity_prefactor: npt.NDArray = (
            self.phase_basic.gravitational_acceleration()
            * self.phase_basic.thermal_expansivity()
            * -self._super_adiabatic_temperature_gradient
        )
        # Viscous velocity
        self._viscous_velocity = (
            velocity_prefactor * self._evaluator.mesh.basic.mixing_length_cubed
        ) / (18 * self.phase_basic.kinematic_viscosity())
        self._viscous_velocity[~self.is_convective] = 0  # Must be super-adiabatic
        # Inviscid velocity
        self._inviscid_velocity = (
            velocity_prefactor * self._evaluator.mesh.basic.mixing_length_squared
        ) / 16
        self._inviscid_velocity[~self.is_convective] = 0  # Must be super-adiabatic
        self._inviscid_velocity[self._is_convective] = np.sqrt(
            self._inviscid_velocity[self._is_convective]
        )
        # Reynolds number
        self._reynolds_number = (
            self._viscous_velocity
            * self._evaluator.mesh.basic.mixing_length
            / self.phase_basic.kinematic_viscosity()
        )
        # Eddy diffusivity
        self._eddy_diffusivity = np.where(
            self.viscous_regime, self._viscous_velocity, self._inviscid_velocity
        )
        self._eddy_diffusivity *= self._evaluator.mesh.basic.mixing_length

        # Heat flux (power per unit area)
        self._heat_flux = np.zeros_like(self.temperature_basic)
        self._mass_flux = np.zeros_like(self.temperature_basic)
        if self._settings.conduction:
            self._heat_flux += self.conductive_heat_flux()
        if self._settings.convection:
            self._heat_flux += self.convective_heat_flux()
        if self._settings.gravitational_separation:
            self._mass_flux += self.gravitational_separation_mass_flux()
        if self._settings.mixing:
            self._mass_flux += self.mixing_mass_flux()
        self._heat_flux += self._mass_flux * self.phase_basic.latent_heat()

        # Heating (power per unit mass)
        self._heating = np.zeros_like(self.temperature_staggered)
        self._heating_radio = np.zeros_like(self.temperature_staggered)
        self._heating_dilatation = np.zeros_like(self.temperature_staggered)
        self._heating_tidal = np.zeros_like(self.temperature_staggered)

        if self._settings.radionuclides:
            self._heating_radio = self.radiogenic_heating(time)
            self._heating += self._heating_radio

        if self._settings.dilatation:
            self._heating_dilatation = self.dilatation_heating()
            self._heating += self._heating_dilatation

        if self._settings.tidal:
            self._heating_tidal = self.tidal_heating()
            self._heating += self._heating_tidal


@dataclass
class Evaluator:
    """Contains classes that evaluate quantities necessary to compute the interior evolution.

    Args:
        _parameters: Parameters

    Attributes:
        boundary_conditions: Boundary conditions
        initial_condition: Initial condition
        mesh: Mesh
        phases: Evaluators for all phases
        radionuclides: Radionuclides
    """

    _parameters: Parameters
    boundary_conditions: BoundaryConditions = field(init=False)
    initial_condition: InitialCondition = field(init=False)
    mesh: Mesh = field(init=False)
    phases: PhaseEvaluatorCollection = field(init=False)

    def __post_init__(self):
        self.mesh = Mesh(self._parameters)
        self.boundary_conditions = BoundaryConditions(self._parameters, self.mesh)
        self.phases = PhaseEvaluatorCollection(self._parameters)
        self.initial_condition = InitialCondition(self._parameters, self.mesh, self.phases)

    @property
    def radionuclides(self) -> list[_Radionuclide]:
        return self._parameters.radionuclides


class Solver:
    """Solves the interior dynamics

    Args:
        filename: Filename of a file with configuration settings
        root: Root path to the flename

    Attributes:
        filename: Filename of a file with configuration settings
        root: Root path to the filename. Defaults to empty
        parameters: Parameters
        evaluator: Evaluator
        state: State
    """

    def __init__(self, param: Parameters):
        logger.info("Creating an Aragog model")
        self.parameters: Parameters = param
        self.evaluator: Evaluator
        self.state: State
        self._solution: OptimizeResult

    @classmethod
    def from_file(cls, filename: str | Path, root: str | Path = Path()) -> Self:
        """Parses a configuration file

        Args:
            filename: Filename
            root: Root of the filename

        Returns:
            Parameters
        """
        configuration_file: Path = Path(root) / Path(filename)
        logger.info("Parsing configuration file = %s", configuration_file)
        parameters: Parameters = Parameters.from_file(configuration_file)

        return cls(parameters)

    def initialize(self) -> None:
        """Initializes the model."""
        logger.info("Initializing %s", self.__class__.__name__)
        self.evaluator = Evaluator(self.parameters)
        self.state = State(self.parameters, self.evaluator)

    def reset(self) -> None:
        """This function initializes the model, while keeping the previous state of the
        PhaseEvaluatorCollection object. This avoids multiple loads of lookup table data
        when running Aragog multiple times.
        """
        logger.info("Resetting %s", self.__class__.__name__)
        # Update the Evaluator object except the phase properties
        self.evaluator.mesh = Mesh(self.parameters)
        self.evaluator.boundary_conditions = BoundaryConditions(self.parameters, self.evaluator.mesh)
        self.evaluator.initial_condition = InitialCondition(self.parameters, self.evaluator.mesh, self.evaluator.phases)
        # Reinstantiate the solver state
        self.state = State(self.parameters, self.evaluator)

    @property
    def temperature_basic(self) -> npt.NDArray:
        """Temperature of the basic mesh in K"""
        return self.evaluator.mesh.quantity_at_basic_nodes(self.temperature_staggered)

    @property
    def temperature_staggered(self) -> npt.NDArray:
        """Temperature of the staggered mesh in K"""
        temperature: npt.NDArray = self.solution.y * self.parameters.scalings.temperature

        return temperature

    @property
    def solution(self) -> OptimizeResult:
        """The solution."""
        return self._solution

    def dTdt(
        self,
        time: npt.NDArray | float,
        temperature: npt.NDArray,
    ) -> npt.NDArray:
        """dT/dt at the staggered nodes

        Args:
            time: Time
            temperature: Temperature at the staggered nodes

        Returns:
            dT/dt at the staggered nodes
        """
        logger.debug("temperature passed into dTdt = %s", temperature)
        # logger.debug("temperature.shape = %s", temperature.shape)
        self.state.update(temperature, time)
        heat_flux: npt.NDArray = self.state.heat_flux
        # logger.debug("heat_flux = %s", heat_flux)
        self.evaluator.boundary_conditions.apply_flux_boundary_conditions(self.state)
        # logger.debug("heat_flux = %s", heat_flux)
        # logger.debug("mesh.basic.area.shape = %s", self.data.mesh.basic.area.shape)

        energy_flux: npt.NDArray = heat_flux * self.evaluator.mesh.basic.area
        # logger.debug("energy_flux size = %s", energy_flux.shape)

        delta_energy_flux: npt.NDArray = np.diff(energy_flux, axis=0)
        # logger.debug("delta_energy_flux size = %s", delta_energy_flux.shape)
        # logger.debug("capacitance = %s", self.state.phase_staggered.capacitance.shape)
        # FIXME: Update capacitance for mixed phase (enthalpy of fusion contribution)
        capacitance: npt.NDArray = (
            self.state.capacitance_staggered() * self.evaluator.mesh.basic.volume
        )

        # Heating rate (dT/dt) from flux divergence (power per unit area)
        dTdt: npt.NDArray = -delta_energy_flux / capacitance
        logger.debug("dTdt (fluxes only) = %s", dTdt)

        # Additional heating rate (dT/dt) from internal heating (power per unit mass)
        dTdt += self.state.heating * (
            self.state.phase_staggered.density() / self.state.capacitance_staggered()
        )

        logger.debug("dTdt (with internal heating) = %s", dTdt)

        return dTdt

    def make_tsurf_event(self):
        """
        Creates a temperature event function for use with an ODE solver to monitor changes 
        in the surface temperature.The event triggers when the change exceeds the 
        threshold, allowing the solver to stop integration.

        Returns:
            The event has the attributes:
                - terminal = True: Integration stops when the event is triggered.
                - direction = -1: Only triggers when the function is decreasing through zero.
        """
        tsurf_initial = [None]

        def tsurf_event(time: float, temperature: npt.NDArray) -> float:
            """
            Event function to detect when surface temperature changes beyond a specified threshold.

            Args:
                time (float): Current time.
                temperature (np.ndarray): Current temperature profile.

            Returns:
                float: The difference between the threshold and the actual change in surface 
                    temperature. When this value crosses zero from above, the event is triggered.
            """
            tsurf_current = temperature[-1] * self.parameters.scalings.temperature
            tsurf_threshold = self.parameters.solver.tsurf_poststep_change * self.parameters.scalings.temperature

            if tsurf_initial[0] is None:
                tsurf_initial[0] = tsurf_current
                return 1.0  
            
            delta = abs(tsurf_current - tsurf_initial[0])

            return tsurf_threshold - delta  

        tsurf_event.terminal = self.parameters.solver.event_triggering
        tsurf_event.direction = -1

        return tsurf_event

    def solve(self) -> None:
        start_time = self.parameters.solver.start_time
        end_time = self.parameters.solver.end_time
        atol = self.parameters.solver.atol
        rtol = self.parameters.solver.rtol

        tsurf_event = self.make_tsurf_event()
  

        self._solution = solve_ivp(
            self.dTdt,
            (start_time, end_time),
            self.evaluator.initial_condition.temperature,
            method="BDF",
            vectorized=True,
            atol=atol,
            rtol=rtol,
            events=[tsurf_event],
         )
        logger.info(self.solution)

        if self._solution.status == 1:
            logger.warning("Integration stopped early due to surface temperature jump.")
            self.stop_early = True

        elif self._solution.status == 0:
            logger.info("Integration completed successfully.")
            self.stop_early = False

        else:
            logger.error("Integration failed with status = %d", self._solution.status)
            logger.error("Message: %s", self._solution.message)
            self.stop_early = True



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""A phase defines the equation of state (EOS) and transport properties."""

from __future__ import annotations

import logging
import sys
from dataclasses import KW_ONLY, Field, InitVar, dataclass, field, fields

import numpy as np
import numpy.typing as npt
from scipy.interpolate import RectBivariateSpline, PchipInterpolator

from aragog.interfaces import (
    MixedPhaseEvaluatorProtocol,
    PhaseEvaluatorABC,
    PhaseEvaluatorProtocol,
    PropertyProtocol,
)
from aragog.parser import Parameters, _PhaseMixedParameters, _PhaseParameters
from aragog.utilities import (
    FloatOrArray,
    combine_properties,
    is_file,
    is_number,
    tanh_weight,
)

if sys.version_info < (3, 12):
    from typing_extensions import override
else:
    from typing import override

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class ConstantProperty(PropertyProtocol):
    """A property with a constant value

    Args:
        name: Name of the property
        value: The constant value

    Attributes:
        name: Name of the property
        value: The constant value
        ndim: Number of dimensions, which is equal to zero for a constant property
    """

    name: str
    _: KW_ONLY
    value: float
    ndim: int = field(init=False, default=0)

    def eval(self) -> float:
        return self.value

    def __call__(self, temperature: npt.NDArray, pressure: npt.NDArray) -> float:
        """Evaluates the property.

        Args:
            temperature: Temperature
            pressure: Pressure
        """
        del temperature
        del pressure
        return self.eval()


@dataclass
class LookupProperty1D(PropertyProtocol):
    """A property from a 1-D lookup

    Args:
        name: Name of the property
        value: A 2-D array, with x values in the first column and y values in the second column.

    Attributes:
        name: Name of the property
        value: A 2-D array
        ndim: Number of dimensions, which is equal to one for a 1-D lookup
    """

    name: str
    _: KW_ONLY
    value: npt.NDArray
    ndim: int = field(init=False, default=1)
    _gradient: npt.NDArray = field(init=False)

    def __post_init__(self):
        # Sort the data to ensure x is increasing
        self.value = self.value[self.value[:, 0].argsort()]
        self._gradient = np.gradient(self.value[:, 1], self.value[:, 0])

    def eval(self, pressure: npt.NDArray) -> npt.NDArray:
        return np.interp(pressure, self.value[:, 0], self.value[:, 1])

    def gradient(self, pressure: npt.NDArray) -> npt.NDArray:
        """Computes the gradient"""
        return np.interp(pressure, self.value[:, 0], self._gradient)

    def __call__(self, temperature: npt.NDArray, pressure: npt.NDArray) -> npt.NDArray:
        del temperature
        return self.eval(pressure)


@dataclass
class LookupProperty2D(PropertyProtocol):
    """A property from a 2-D lookup

    Args:
        name: Name of the property
        value: The 2-D array

    Attributes:
        name: Name of the property
        value: The 2-D array
        ndim: Number of dimensions, which is equal to two for a 2-D lookup
    """

    name: str
    _: KW_ONLY
    value: npt.NDArray
    ndim: int = field(init=False, default=2)
    _lookup: RectBivariateSpline = field(init=False)

    def __post_init__(self):
        # Prepare data for spline
        x_values, y_values, z_values = self.prepare_data_for_spline(self.value)
        self._lookup = RectBivariateSpline(x_values, y_values, z_values, kx=1, ky=1, s=0)

    def prepare_data_for_spline(self, data):
        """Ensure your data is on a regular grid for RectBivariateSpline"""
        # Extract x, y, and z values
        x_values = np.unique(data[:, 0])  # Unique pressure values
        y_values = np.unique(data[:, 1])  # Unique temperature values

        # Create a grid for z values
        z_values = np.full((x_values.size, y_values.size), np.nan)

        # Find the indices of the x and y values in the unique arrays
        x_indices = np.searchsorted(x_values, data[:, 0])
        y_indices = np.searchsorted(y_values, data[:, 1])

        # Fill the z_values grid
        z_values[x_indices, y_indices] = data[:, 2]

        return x_values, y_values, z_values

    def eval(self, temperature: npt.NDArray, pressure: npt.NDArray) -> npt.NDArray:
        return self._lookup(pressure, temperature, grid=False)

    def __call__(self, temperature: npt.NDArray, pressure: npt.NDArray) -> npt.NDArray:
        return self.eval(temperature, pressure)


class SinglePhaseEvaluator(PhaseEvaluatorABC):
    """Contains the objects to evaluate the EOS and transport properties of a phase.

    Args:
        settings: Phase parameters
        gravitational_acceleration: PropertyProtocol
    """

    # For typing
    _density: PropertyProtocol
    _gravitational_acceleration: PropertyProtocol
    _heat_capacity: PropertyProtocol
    _melt_fraction: ConstantProperty
    _thermal_conductivity: PropertyProtocol
    _thermal_expansivity: PropertyProtocol
    _viscosity: PropertyProtocol

    def __init__(self, settings: _PhaseParameters, gravitational_acceleration: PropertyProtocol):
        self._settings: _PhaseParameters = settings
        cls_fields: tuple[Field, ...] = fields(self._settings)
        for field_ in cls_fields:
            name: str = field_.name
            private_name: str = f"_{name}"
            value = getattr(self._settings, field_.name)

            if is_number(value):
                # Numbers have already been scaled by the parser
                setattr(self, private_name, ConstantProperty(name=name, value=value))

            elif is_file(value):
                with open(value, encoding="utf-8") as infile:
                    logger.debug("%s is a file = %s", name, value)
                    header = infile.readline()
                    col_names = header[1:].split()
                value_array: npt.NDArray = np.loadtxt(value, ndmin=2)
                logger.debug("before scaling, value_array = %s", value_array)
                # Scale lookup data
                for nn, col_name in enumerate(col_names):
                    logger.info("Scaling %s from %s", col_name, value)
                    value_array[:, nn] /= getattr(self._settings.scalings_, col_name)
                logger.debug("after scaling, value_array = %s", value_array)
                ndim = value_array.shape[1]
                logger.debug("ndim = %d", ndim)
                if ndim == 2:
                    setattr(self, private_name, LookupProperty1D(name=name, value=value_array))
                elif ndim == 3:
                    setattr(self, private_name, LookupProperty2D(name=name, value=value_array))
                else:
                    raise ValueError(f"Lookup data must have 2 or 3 dimensions, not {ndim}")
            else:
                logger.info("Cannot interpret value (%s): not a number or a file", value)

        self._gravitational_acceleration = gravitational_acceleration

    @override
    def density(self) -> FloatOrArray:
        return self._density(self.temperature, self.pressure)

    @override
    def gravitational_acceleration(self) -> FloatOrArray:
        return self._gravitational_acceleration(self.temperature, self.pressure)

    @override
    def heat_capacity(self) -> FloatOrArray:
        return self._heat_capacity(self.temperature, self.pressure)

    @override
    def melt_fraction(self) -> float:
        return self._melt_fraction(self.temperature, self.pressure)
    
    @override
    def latent_heat(self) -> float:
        return 0.

    @override
    def thermal_conductivity(self) -> FloatOrArray:
        return self._thermal_conductivity(self.temperature, self.pressure)

    @override
    def thermal_expansivity(self) -> FloatOrArray:
        return self._thermal_expansivity(self.temperature, self.pressure)

    @override
    def viscosity(self) -> FloatOrArray:
        return self._viscosity(self.temperature, self.pressure)

    @override
    def relative_velocity(self) -> float:
        return 0.

    @override
    def delta_specific_volume(self) -> FloatOrArray:
        return 0.0

class MixedPhaseEvaluator(PhaseEvaluatorABC):
    """Evaluates the EOS and transport properties of a mixed phase.

    This only computes quantities within the mixed phase region between the solidus and the
    liquidus. Computing quantities outside of this region will give incorrect results.

    Args:
        parameters: Parameters

    Attributes:
        settings: Mixed phase parameters
    """

    def __init__(self, parameters: Parameters):
        gravitational_acceleration: PropertyProtocol = setup_gravitational_acceleration(parameters)
        self.settings: _PhaseMixedParameters = parameters.phase_mixed
        self._liquid: PhaseEvaluatorProtocol = SinglePhaseEvaluator(
            parameters.phase_liquid, gravitational_acceleration
        )
        self._solid: PhaseEvaluatorProtocol = SinglePhaseEvaluator(
            parameters.phase_solid, gravitational_acceleration
        )
        self._solidus: LookupProperty1D = self._get_melting_curve_lookup(
            "solidus", self.settings.solidus
        )
        self._liquidus: LookupProperty1D = self._get_melting_curve_lookup(
            "liquidus", self.settings.liquidus
        )
        self._grain_size: float = self.settings.grain_size
        self._latent_heat: float = self.settings.latent_heat_of_fusion

    @override
    def set_pressure(self, pressure: npt.NDArray) -> None:
        """Sets pressure and updates quantities that only depend on pressure"""
        super().set_pressure(pressure)
        # Sets the temperature of the solid and liquid phases to the appropriate melting curve in
        # order to evaluate mixed properties
        self._solid.set_temperature(self.solidus())
        self._solid.set_pressure(pressure)
        self._liquid.set_temperature(self.liquidus())
        self._liquid.set_pressure(pressure)
        self._delta_density = self._solid.density() - self._liquid.density()
        self._delta_specific_volume = 1.0/self._liquid.density() - 1.0/self._solid.density()
        self._delta_fusion = self.liquidus() - self.solidus()
        # Heat capacity of the mixed phase :cite:p:`{Equation 4,}SOLO07`
        self._heat_capacity = self.latent_heat() / self.delta_fusion()

    @override
    def update(self):
        # Melt fraction without clipping
        # phi<0 for the solid, 0<phi<1 for the mixed phase, and phi>1 for the melt
        self._melt_fraction_no_clip = (self.temperature - self.solidus()) / self.delta_fusion()
        logger.debug("_melt_fraction_no_clip = %s", self.melt_fraction_no_clip())
        self._melt_fraction = np.clip(self._melt_fraction_no_clip, 0, 1)

        # Mixed density by volume additivity
        self._density = combine_properties(
            self.melt_fraction(), 1 / self._liquid.density(), 1 / self._solid.density()
        )
        self._density = 1 / self._density

        # Porosity
        epsilon = 1e-10  # Used to avoid division by zero when dividing by the delta density for higher mass planets
        self._porosity = (self._solid.density() - self.density()) / (self.delta_density() + epsilon)

        # Relative velocity between melt and solid
        self._relative_velocity = self._get_relative_velocity()

        # Thermal conductivity
        self._thermal_conductivity = combine_properties(
            self.melt_fraction(),
            self._liquid.thermal_conductivity(),
            self._solid.thermal_conductivity(),
        )

        # Thermal expansivity :cite:p:`{Equation 3,}SOLO07`
        # The first term in :cite:t:`{Equation 3,}SOLO07` is not included because it is small
        # compared to the latent heat term :cite:p:`{Equation 33,}SS93`.
        self._thermal_expansivity = self.delta_density() / self.delta_fusion() / self.density()

        # Viscosity
        weight: npt.NDArray = tanh_weight(
            self.melt_fraction(),
            self.settings.rheological_transition_melt_fraction,
            self.settings.rheological_transition_width,
        )
        log10_viscosity: npt.NDArray = combine_properties(
            weight, np.log10(self._liquid.viscosity()), np.log10(self._solid.viscosity())
        )
        self._viscosity = 10**log10_viscosity

    def delta_density(self) -> FloatOrArray:
        return self._delta_density

    def delta_fusion(self) -> npt.NDArray:
        return self._delta_fusion

    @override
    def density(self) -> npt.NDArray:
        return self._density

    @override
    def gravitational_acceleration(self) -> FloatOrArray:
        return self._solid.gravitational_acceleration()

    @override
    def heat_capacity(self) -> FloatOrArray:
        """Heat capacity of the mixed phase :cite:p:`{Equation 4,}SOLO07`"""
        return self._heat_capacity

    def liquidus(self) -> npt.NDArray:
        """Liquidus"""
        liquidus: npt.NDArray = self._liquidus.eval(self.pressure)
        logger.debug("liquidus = %s", liquidus)

        return liquidus

    def liquidus_gradient(self) -> npt.NDArray:
        """Liquidus gradient"""
        return self._liquidus.gradient(self.pressure)

    def melt_fraction_no_clip(self) -> npt.NDArray:
        """Melt fraction without clipping"""
        return self._melt_fraction_no_clip

    @override
    def melt_fraction(self) -> npt.NDArray:
        """Melt fraction of the mixed phase

        The melt fraction is always between zero and one.
        """
        return self._melt_fraction

    @override
    def latent_heat(self) -> float:
        """Latent heat of fusion"""
        return self._latent_heat

    def porosity(self) -> npt.NDArray:
        """Porosity of the mixed phase, that is the volume fraction occupied by the melt"""
        return self._porosity

    def solidus(self) -> npt.NDArray:
        """Solidus"""
        solidus: npt.NDArray = self._solidus.eval(self.pressure)
        logger.debug("solidus = %s", solidus)

        return solidus

    def solidus_gradient(self) -> npt.NDArray:
        """Solidus gradient"""
        return self._solidus.gradient(self.pressure)

    @override
    def thermal_conductivity(self) -> npt.NDArray:
        return self._thermal_conductivity

    @override
    def thermal_expansivity(self) -> npt.NDArray:
        return self._thermal_expansivity

    @override
    def viscosity(self) -> npt.NDArray:
        return self._viscosity

    @override
    def relative_velocity(self) -> npt.NDArray:
        """Relative velocity between melt and solid"""
        return self._relative_velocity

    def _get_relative_velocity(self) -> npt.NDArray:
        """Compute relative velocity"""
        dv = (
            - self.delta_density()
            * self.gravitational_acceleration()
            * self._permeability()
            / self._liquid.viscosity()
        )
        return dv

    @override
    def delta_specific_volume(self) -> npt.NDArray:
        """Difference of specific volume between melt and solid"""
        return self._delta_specific_volume

    def _permeability(self) -> npt.NDArray:

        # RumpfGupte regime (default)
        permeability = self._permeability_rumpf_gupte()

        # Stokes regime
        permeability = np.where(
            self._porosity > 0.771462,
            self._permeability_stokes(),
            permeability
            )

        # Blake-Kozeny-Carman regime
        permeability = np.where(
            self._porosity < 0.0769452,
            self._permeability_blake_kozeny_carman(),
            permeability
            )
        
        return permeability

    def _permeability_stokes(self) -> npt.NDArray:
        """Permeability for Stokes flow in the mixed phase"""
        permeability = 2./9.*self._grain_size**2
        return permeability

    def _permeability_blake_kozeny_carman(self) -> npt.NDArray:
        """Permeability for Blake-Kozeny-Carman flow in the mixed phase"""
        permeability = (
            0.001
            * self._grain_size**2
            * self._porosity**2
            / (1-self._porosity)**2
        )
        return permeability

    def _permeability_rumpf_gupte(self) -> npt.NDArray:
        """Permeability for Rumpf-Gupte flow in the mixed phase"""
        permeability = (
            5./7.
            * self._grain_size**2
            * self._porosity**4.5
        )
        return permeability

    def _get_melting_curve_lookup(self, name: str, value: str) -> LookupProperty1D:
        with open(value, encoding="utf-8") as infile:
            header = infile.readline()
            col_names = header[1:].split()
        value_array: npt.NDArray = np.loadtxt(value, ndmin=2)
        logger.debug("before scaling, value_array = %s", value_array)
        for nn, col_name in enumerate(col_names):
            logger.info("Scaling %s from %s", col_name, value)
            value_array[:, nn] /= getattr(self.settings.scalings_, col_name)

        return LookupProperty1D(name=name, value=value_array)


class CompositePhaseEvaluator(PhaseEvaluatorABC):
    """Evaluates the EOS and transport properties of a composite phase.

    This combines the single phase evaluators for the liquid and solid regions with the mixed phase
    evaluator for the mixed phase region. This ensure that the phase properties are computed
    correctly for all temperatures and pressures.

    Args:
        parameters: Parameters
    """

    def __init__(self, parameters: Parameters):
        gravitational_acceleration: PropertyProtocol = setup_gravitational_acceleration(parameters)
        self._liquid: PhaseEvaluatorProtocol = SinglePhaseEvaluator(
            parameters.phase_liquid, gravitational_acceleration
        )
        self._solid: PhaseEvaluatorProtocol = SinglePhaseEvaluator(
            parameters.phase_solid, gravitational_acceleration
        )
        self._mixed: MixedPhaseEvaluator = MixedPhaseEvaluator(parameters)

    @override
    def set_temperature(self, temperature: npt.NDArray) -> None:
        super().set_temperature(temperature)
        self._solid.set_temperature(temperature)
        self._liquid.set_temperature(temperature)
        self._mixed.set_temperature(temperature)

    @override
    def set_pressure(self, pressure: npt.NDArray) -> None:
        """Sets pressure and updates quantities that only depend on pressure"""
        super().set_pressure(pressure)
        self._solid.set_pressure(pressure)
        self._liquid.set_pressure(pressure)
        self._mixed.set_pressure(pressure)

    @override
    def update(self) -> None:
        self._mixed.update()
        self._set_blending_and_masks()
        self._density = self._get_composite("density")
        self._heat_capacity = self._get_composite("heat_capacity")
        self._thermal_conductivity = self._get_composite("thermal_conductivity")
        self._thermal_expansivity = self._get_composite("thermal_expansivity")
        self._dTdPs = self._get_composite("dTdPs")
        self._dTdrs = self._get_composite("dTdrs")
        self._relative_velocity = self._get_composite("relative_velocity")
        self._delta_specific_volume = self._get_composite("delta_specific_volume")

        name: str = "viscosity"
        log10_mixed_phase: npt.NDArray = np.log10(getattr(self._mixed, name)())
        single_phase: npt.NDArray = np.empty_like(self._blending_factor)
        try:
            single_phase[self._liquid_mask] = getattr(self._liquid, name)()[self._liquid_mask]
        except (IndexError, TypeError):
            single_phase[self._liquid_mask] = getattr(self._liquid, name)()
        try:
            single_phase[self._solid_mask] = getattr(self._solid, name)()[self._solid_mask]
        except (IndexError, TypeError):
            single_phase[self._solid_mask] = getattr(self._solid, name)()
        log10_single_phase: npt.NDArray = np.log10(single_phase)
        self._viscosity = combine_properties(
            self._blending_factor, log10_mixed_phase, log10_single_phase
        )
        self._viscosity = 10**self._viscosity

    @override
    def density(self) -> npt.NDArray:
        return self._density

    @override
    def dTdPs(self) -> npt.NDArray:
        return self._dTdPs

    @override
    def dTdrs(self) -> npt.NDArray:
        return self._dTdrs

    @override
    def gravitational_acceleration(self) -> FloatOrArray:
        return self._mixed.gravitational_acceleration()

    @override
    def heat_capacity(self) -> npt.NDArray:
        """Heat capacity"""
        return self._heat_capacity

    def liquidus(self) -> npt.NDArray:
        return self._mixed.liquidus()

    def liquidus_gradient(self) -> npt.NDArray:
        return self._mixed.liquidus_gradient()

    @override
    def melt_fraction(self) -> npt.NDArray:
        """Melt fraction"""
        return self._mixed.melt_fraction()

    @override
    def latent_heat(self) -> FloatOrArray:
        """Latent heat of fusion"""
        return self._mixed.latent_heat()

    def solidus(self) -> npt.NDArray:
        return self._mixed.solidus()

    def solidus_gradient(self) -> npt.NDArray:
        return self._mixed.solidus_gradient()

    @override
    def thermal_conductivity(self) -> npt.NDArray:
        """Thermal conductivity"""
        return self._thermal_conductivity

    @override
    def thermal_expansivity(self) -> npt.NDArray:
        """Thermal expansivity"""
        return self._thermal_expansivity

    def viscosity(self) -> npt.NDArray:
        """Viscosity"""
        return self._viscosity

    @override
    def relative_velocity(self) -> npt.NDArray:
        """Relative velocity between melt and solid"""
        return self._relative_velocity

    @override
    def delta_specific_volume(self) -> npt.NDArray:
        """Difference of specific volume between melt and solid"""
        return self._delta_specific_volume

    def _set_blending_and_masks(self) -> None:
        """Sets blending and masks."""

        phase_transition_width: float = self._mixed.settings.phase_transition_width
        melt_fraction_no_clip: npt.NDArray = self._mixed.melt_fraction_no_clip()

        if phase_transition_width == 0.0:
            blending_factor: npt.NDArray = np.where(
                ((melt_fraction_no_clip < 0.0) | (melt_fraction_no_clip > 1.0)),
                0,
                1,
            )
        else:
            blending_liquid: npt.NDArray = 1.0 - tanh_weight(
                melt_fraction_no_clip, 1.0, phase_transition_width
            )
            blending_solid: npt.NDArray = tanh_weight(
                melt_fraction_no_clip, 0.0, phase_transition_width
            )
            blending_factor = np.where(
                melt_fraction_no_clip > 0.5, blending_liquid, blending_solid
            )

        self._blending_factor = blending_factor
        logger.debug("_blending_factor = %s", self._blending_factor)
        self._liquid_mask = melt_fraction_no_clip > 0.5
        logger.debug("_liquid_mask = %s", self._liquid_mask)
        self._solid_mask = ~self._liquid_mask
        logger.debug("_solid_mask = %s", self._solid_mask)

    def _get_composite(self, property_name: str) -> npt.NDArray:
        """Evaluates the composite property"""
        mixed_phase: npt.NDArray = getattr(self._mixed, property_name)()
        single_phase: npt.NDArray = np.empty_like(self._blending_factor)
        logger.debug("single_phase = %s", single_phase)
        logger.debug("_liquid_mask = %s", self._liquid_mask)
        logger.debug("_solid_mask = %s", self._solid_mask)
        test = getattr(self._liquid, property_name)()
        logger.debug("test = %s", test)

        # logger.debug(self.temperature.shape)
        # logger.debug(self.pressure.shape)
        # logger.debug(mixed_phase.shape)
        # logger.debug(single_phase.shape)

        # TODO: This is ugly.  Clean up logic.
        try:
            single_phase[self._liquid_mask] = getattr(self._liquid, property_name)()[
                self._liquid_mask
            ]
        except (IndexError, TypeError):
            single_phase[self._liquid_mask] = getattr(self._liquid, property_name)()
        try:
            single_phase[self._solid_mask] = getattr(self._solid, property_name)()[
                self._solid_mask
            ]
        except (IndexError, TypeError):
            single_phase[self._solid_mask] = getattr(self._solid, property_name)()

        combined: npt.NDArray = combine_properties(
            self._blending_factor, mixed_phase, single_phase
        )

        return combined


@dataclass
class PhaseEvaluatorCollection:
    """A collection of phase evaluators

    Creates the phase evaluators and selects the active phase based on configuration data.

    Args:
        parameters: Parameters

    Attributes:
        liquid: Liquid evaluator
        solid: Solid evaluator
        mixed: Mixed evaluator
        composite: Composite evaluator
        active: The active evaluator, which is defined by configuration data
    """

    parameters: InitVar[Parameters]
    liquid: PhaseEvaluatorProtocol = field(init=False)
    solid: PhaseEvaluatorProtocol = field(init=False)
    mixed: MixedPhaseEvaluatorProtocol = field(init=False)
    composite: MixedPhaseEvaluatorProtocol = field(init=False)
    active: PhaseEvaluatorProtocol = field(init=False)

    def __post_init__(self, parameters: Parameters):
        gravitational_acceleration: PropertyProtocol = setup_gravitational_acceleration(parameters)
        self.liquid = SinglePhaseEvaluator(parameters.phase_liquid, gravitational_acceleration)
        self.solid = SinglePhaseEvaluator(parameters.phase_solid, gravitational_acceleration)
        self.mixed = MixedPhaseEvaluator(parameters)
        self.composite = CompositePhaseEvaluator(parameters)

        # Configuration data defines which phase to use for the model.
        phase_to_use: str = parameters.phase_mixed.phase

        if phase_to_use == "liquid":
            self.active = self.liquid
        elif phase_to_use == "solid":
            self.active = self.solid
        # Allowing selection of self.mixed doesn't really make sense because it will probably give
        # crazy results outside the mixed phase region. Hence just use composite.
        elif phase_to_use == "mixed" or phase_to_use == "composite":
            self.active = self.composite
        else:
            raise ValueError(f"Phase = {phase_to_use} is not a valid selection")

def setup_gravitational_acceleration(parameters: Parameters):
    """Sets up the gravitational acceleration property.

    Args:
        parameters: Parameters

    Returns:
        gravitational_acceleration: PropertyProtocol
    """
    if parameters.mesh.eos_method == 1:
        # AdamsWilliamson EOS method = constant gravitational acceleration
        gravitational_acceleration = ConstantProperty(
            "gravitational_acceleration", value=parameters.mesh.gravitational_acceleration
        )
    elif parameters.mesh.eos_method == 2:
        # User defined EOS method = space dependent gravitational acceleration (1D lookup)
        # Re-interpolate EOS-data to match standard lookup pressure grid
        interp_gravity = PchipInterpolator(
            np.flip(parameters.mesh.eos_pressure),
            np.flip(parameters.mesh.eos_gravity))
        num_points: int = 138
        max_pressure: float = 1.37e11 / parameters.scalings.pressure
        lookup_data: npt.NDArray = np.zeros((num_points, 2), dtype=float)
        lookup_data[:,0] = np.linspace(0.0, max_pressure, num=num_points)
        lookup_data[:,1] = interp_gravity(lookup_data[:,0])
        gravitational_acceleration = LookupProperty1D(
            "gravitational_acceleration",
            value=lookup_data,
        )
    else:
        raise ValueError(f"EOS method = {parameters.mesh.eos_method} is not a valid selection")

    return gravitational_acceleration


#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Package level variables and initialises the package logger"""

from __future__ import annotations

__version__: str = "0.2.7-alpha"

import importlib.resources
import logging
import os
from importlib.abc import Traversable

CFG_DATA: Traversable = importlib.resources.files(f"{__package__}.cfg")

# Creates the package logger.
# https://docs.python.org/3/howto/logging.html#library-config
logger: logging.Logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def complex_formatter() -> logging.Formatter:
    """Complex formatter for logging

    Returns:
        Formatter for logging
    """
    fmt: str = "[%(asctime)s - %(name)-30s - %(lineno)03d - %(levelname)-9s - %(funcName)s()]"
    fmt += " - %(message)s"
    datefmt: str = "%Y-%m-%d %H:%M:%S"
    formatter: logging.Formatter = logging.Formatter(fmt, datefmt=datefmt)

    return formatter


def simple_formatter() -> logging.Formatter:
    """Simple formatter for logging

    Returns:
        Formatter for logging
    """
    fmt: str = "[%(asctime)s - %(name)-30s - %(levelname)-9s] - %(message)s"
    datefmt: str = "%H:%M:%S"
    formatter: logging.Formatter = logging.Formatter(fmt, datefmt=datefmt)

    return formatter


def debug_logger() -> logging.Logger:
    """Sets up debug logging to the console.

    Returns:
        A logger
    """
    package_logger: logging.Logger = logging.getLogger(__name__)
    package_logger.setLevel(logging.DEBUG)
    logger.handlers = []
    console_handler: logging.Handler = logging.StreamHandler()
    console_formatter: logging.Formatter = simple_formatter()
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    return package_logger


def aragog_file_logger(
    console_level=logging.INFO, file_level=logging.DEBUG, log_dir=os.getcwd()
) -> logging.Logger:
    """Sets up console logging and file logging according to arguments.

    Returns:
        A logger
    """
    # Console logger
    package_logger: logging.Logger = logging.getLogger(__name__)
    package_logger.setLevel(file_level)
    package_logger.handlers = []
    console_handler: logging.Handler = logging.StreamHandler()
    console_formatter: logging.Formatter = simple_formatter()
    console_handler.setFormatter(console_formatter)
    console_handler.setLevel(console_level)
    logger.addHandler(console_handler)
    # File logger
    log_file = os.path.join(log_dir, f"{__package__}.log")
    file_handler: logging.Handler = logging.FileHandler(log_file)
    file_formatter: logging.Formatter = complex_formatter()
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)
    package_logger.addHandler(file_handler)

    return package_logger


# Expose public API so pylint: disable = C0413
from aragog.data import DownloadLookupTableData
from aragog.output import Output
from aragog.solver import Solver



import logging
import os
from pathlib import Path
from time import sleep
import subprocess as sp

import platformdirs
from osfclient.api import OSF

logger: logging.Logger = logging.getLogger(__name__)

FWL_DATA_DIR = Path(os.environ.get("FWL_DATA", platformdirs.user_data_dir("fwl_data")))

logger.info(f"FWL data location: {FWL_DATA_DIR}")

# project ID of the lookup data folder in the OSF
project_id = "phsxf"

basic_list = (
    "1TPa-dK09-elec-free/MgSiO3_Wolf_Bower_2018",
    "Melting_curves/Wolf_Bower+2018",
    )

full_list = (
    "1TPa-dK09-elec-free/MgSiO3_Wolf_Bower_2018",
    "1TPa-dK09-elec-free/MgSiO3_Wolf_Bower_2018_400GPa",
    "Melting_curves/Monteux+600",
    "Melting_curves/Monteux-600",
    "Melting_curves/Wolf_Bower+2018",
    )

def get_zenodo_record(folder: str) -> str | None:
    """
    Get Zenodo record ID for a given folder.

    Inputs :
        - folder : str
            Folder name to get the Zenodo record ID for

    Returns :
        - str | None : Zenodo record ID or None if not found
    """
    zenodo_map = {
        '1TPa-dK09-elec-free/MgSiO3_Wolf_Bower_2018': '15877374',
        '1TPa-dK09-elec-free/MgSiO3_Wolf_Bower_2018_400GPa': '15877424',
        'Melting_curves/Monteux+600': '15728091',
        'Melting_curves/Monteux-600': '15728138',
        'Melting_curves/Wolf_Bower+2018': '15728072',
    }
    return zenodo_map.get(folder, None)

def download_zenodo_folder(folder: str, data_dir: Path):
    """
    Download a specific Zenodo record into specified folder

    Inputs :
        - folder : str
            Folder name to download
        - folder_dir : Path
            local repository where data are saved
    """

    folder_dir = data_dir / folder
    folder_dir.mkdir(parents=True)
    zenodo_id = get_zenodo_record(folder)
    cmd = [
            "zenodo_get", zenodo_id,
            "-o", folder_dir
        ]
    out = os.path.join(GetFWLData(), "zenodo.log")
    logger.debug("    logging to %s"%out)
    with open(out,'w') as hdl:
        sp.run(cmd, check=True, stdout=hdl, stderr=hdl)

def download_OSF_folder(*, storage, folders: list[str], data_dir: Path):
    """
    Download a specific folder in the OSF repository

    Inputs :
        - storage : OSF storage name
        - folders : folder names to download
        - data_dir : local repository where data are saved
    """
    for file in storage.files:
        for folder in folders:
            if not file.path[1:].startswith(folder):
                continue
            parts = file.path.split("/")[1:]
            target = Path(data_dir, *parts)
            target.parent.mkdir(parents=True, exist_ok=True)
            logger.info(f"Downloading {file.path}...")
            with open(target, "wb") as f:
                file.write_to(f)
            break

def GetFWLData() -> Path:
    """
    Get path to FWL data directory on the disk
    """
    return Path(FWL_DATA_DIR).absolute()

def DownloadLookupTableData(fname: str = ""):
    """
    Download lookup table data

    Inputs :
        - fname (optional) :    folder name, i.e. "1TPa-dK09-elec-free/MgSiO3_Wolf_Bower_2018"
                                if not provided download all the folder list
    """

    osf = OSF()
    project = osf.project(project_id)
    storage = project.storage("osfstorage")

    data_dir = GetFWLData() / "interior_lookup_tables"
    data_dir.mkdir(parents=True, exist_ok=True)

    # If no folder specified download all basic list
    if not fname:
        folder_list = basic_list
    elif fname in full_list:
        folder_list = [fname]
    else:
        raise ValueError(f"Unrecognised folder name: {fname}")

    for folder in folder_list:
        folder_dir = data_dir / folder
        max_tries = 2 # Maximum download attempts, could be a function argument

        if not folder_dir.exists():
            logger.info(f"Downloading interior lookup table data to {data_dir}")
            for i in range(max_tries):
                logger.info(f"Attempt {i + 1} of {max_tries}")
                success = False

                try:
                    download_zenodo_folder(folder = folder, data_dir=data_dir)
                    success = True
                except RuntimeError as e:
                    logger.error(f"Zenodo download failed: {e}")
                    folder_dir.rmdir()

                if not success:
                    try:
                        download_OSF_folder(storage=storage, folders=folder, data_dir=data_dir)
                        success = True
                    except RuntimeError as e:
                        logger.error(f"OSF download failed: {e}")

                if success:
                    break

                if i < max_tries - 1:
                    logger.info("Retrying download...")
                    sleep(5) # Wait 5 seconds before retrying
                else:
                    logger.error("Max retries reached. Download failed.")

    return



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Aragog.
#
# Aragog is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Aragog is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Aragog. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Mesh"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import cached_property
from scipy.interpolate import PchipInterpolator

import numpy as np
import numpy.typing as npt

from aragog.parser import Parameters, _MeshParameters
from aragog.utilities import FloatOrArray, is_monotonic_increasing

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class FixedMesh:
    """A fixed mesh

    Args:
        settings: Mesh parameters
        radii: Radii of the mesh
        mass_radii: Mass coordinates of the mesh
        outer_boundary: Outer boundary for computing depth below the surface
        inner_boundary: Inner boundary for computing height above the base

    Attributes:
        settings: Mesh parameters
        radii: Radii of the mesh
        mass_radii: Mass coordinates of the mesh
        outer_boundary: Outer boundary for computing depth below the surface
        inner_boundary: Inner boundary for computing height above the base
        area: Surface area
        delta_mesh: Delta radii in mass coordinates
        depth: Depth below the outer boundary
        height: Height above the inner boundary
        mixing_length: Mixing length
        mixing_length_squared: Mixing length squared
        mixing_length_cubed: Mixing length cubed
        number_of_nodes: Number of nodes
        volume: Volume of the spherical shells defined between neighbouring radii
        total_volume: Total volume
    """

    settings: _MeshParameters
    radii: npt.NDArray
    mass_radii: npt.NDArray
    outer_boundary: float
    inner_boundary: float

    def __post_init__(self):
        if not is_monotonic_increasing(self.radii):
            msg: str = "Mesh must be monotonically increasing"
            logger.error(msg)
            raise ValueError(msg)

    @cached_property
    def area(self) -> npt.NDArray:
        """Area"""
        return 4 * np.pi * np.square(self.radii)

    @cached_property
    def delta_mesh(self) -> npt.NDArray:
        return np.diff(self.mass_radii, axis=0)

    @cached_property
    def depth(self) -> npt.NDArray:
        return self.outer_boundary - self.radii

    @cached_property
    def height(self) -> npt.NDArray:
        return self.radii - self.inner_boundary

    @cached_property
    def _mesh_cubed(self) -> npt.NDArray:
        return np.power(self.radii, 3)

    @cached_property
    def mixing_length(self) -> npt.NDArray:
        if self.settings.mixing_length_profile == "nearest_boundary":
            logger.debug("Set mixing length profile to nearest boundary")
            mixing_length = np.minimum(
                self.outer_boundary - self.radii, self.radii - self.inner_boundary
            )
        elif self.settings.mixing_length_profile == "constant":
            logger.debug("Set mixing length profile to constant")
            assert self.outer_boundary is not None
            assert self.inner_boundary is not None
            mixing_length = (
                np.ones_like(self.radii) * 0.25 * (self.outer_boundary - self.inner_boundary)
            )
        else:
            msg: str = f"Mixing length profile = {self.settings.mixing_length_profile} is unknown"
            raise ValueError(msg)

        return mixing_length

    @cached_property
    def mixing_length_cubed(self) -> npt.NDArray:
        return np.power(self.mixing_length, 3)

    @cached_property
    def mixing_length_squared(self) -> npt.NDArray:
        return np.square(self.mixing_length)

    @cached_property
    def number_of_nodes(self) -> int:
        return self.radii.size

    @cached_property
    def volume(self) -> npt.NDArray:
        volume: npt.NDArray = 4 / 3 * np.pi * (self._mesh_cubed[1:] - self._mesh_cubed[:-1])

        return volume

    @cached_property
    def total_volume(self) -> float:
        return 4 / 3 * np.pi * float(self._mesh_cubed[-1] - self._mesh_cubed[0])


class Mesh:
    """A staggered mesh.

    The basic mesh is used for the flux calculations and the staggered mesh is used for the volume
    calculations.

    Args:
        parameters: Parameters
    """

    eos: EOS = field(init=False)

    def __init__(self, parameters: Parameters):

        # STEP 1: Set up the basic mesh
        self.settings: _MeshParameters = parameters.mesh
        basic_coordinates: npt.NDArray = self.get_constant_spacing()
        if self.settings.eos_method == 1:
            self.eos = AdamsWilliamsonEOS(
                self.settings, basic_coordinates
            )
        elif self.settings.eos_method == 2:
            self.eos = UserDefinedEOS(
                self.settings, basic_coordinates
            )
        else:
            msg: str = (f"Unknown method to initialize Equation of State")
            raise ValueError(msg)
        if parameters.mesh.mass_coordinates:
            self._planet_density: float = self.get_planet_density(basic_coordinates)
            basic_mass_coordinates: npt.NDArray = (
                self.get_basic_mass_coordinates_from_spatial_coordinates(basic_coordinates))
            logger.debug("Basic mass coordinates = %s", basic_mass_coordinates)
        else:
            basic_mass_coordinates = basic_coordinates
        self.basic: FixedMesh = FixedMesh(
            self.settings,
            basic_coordinates,
            basic_mass_coordinates,
            np.max(basic_coordinates),
            np.min(basic_coordinates)
        )

        # STEP 2: Set up the staggered mesh
        staggered_mass_coordinates: npt.NDArray = (
            self.basic.mass_radii[:-1] + 0.5 * self.basic.delta_mesh)
        if parameters.mesh.mass_coordinates:
            staggered_coordinates: npt.NDArray = (
                self.get_staggered_spatial_coordinates_from_mass_coordinates(staggered_mass_coordinates))
        else:
            staggered_coordinates = staggered_mass_coordinates
        self.staggered: FixedMesh = FixedMesh(
            self.settings,
            staggered_coordinates,
            staggered_mass_coordinates,
            self.basic.outer_boundary,
            self.basic.inner_boundary,
        )
        self.eos.set_staggered_pressure(self.staggered.radii)

        # STEP 3: Set up the transform matrices
        if parameters.mesh.mass_coordinates:
            self._dxidr: npt.NDArray = self.get_dxidr_basic()
        else:
            self._dxidr: npt.NDArray = np.ones_like(self.basic.radii)
        self._d_dr_transform: npt.NDArray = self._get_d_dr_transform_matrix()
        self._quantity_transform: npt.NDArray = self._get_quantity_transform_matrix()

    @property
    def dxidr(self) -> npt.NDArray:
        """dxi/dr at basic nodes"""
        return self._dxidr

    @cached_property
    def staggered_effective_density(self) -> npt.NDArray:
        return self.eos.staggered_effective_density

    @cached_property
    def basic_pressure(self) -> npt.NDArray:
        return self.eos.basic_pressure

    @cached_property
    def staggered_pressure(self) -> npt.NDArray:
        return self.eos.staggered_pressure

    def get_planet_density(self, basic_coordinates: npt.NDArray) -> float:
        """Computes the planet density.

        Args:
            Basic spatial coordinates

        Returns:
            Planet effective density
        """
        core_mass = self.settings.core_density *  np.power(basic_coordinates[0,0], 3.0)
        basic_volumes = (np.power(basic_coordinates[1:,0],3.0)
            - np.power(basic_coordinates[:-1,0],3.0))
        mantle_mass = np.sum(
            self.staggered_effective_density[:,0] * basic_volumes
        )
        planet_density = (
            (core_mass + mantle_mass) / (np.power(basic_coordinates[-1,0],3.0)))
        return planet_density

    def get_basic_mass_coordinates_from_spatial_coordinates(self, basic_coordinates: npt.NDArray) -> npt.NDArray:
        """Computes the basic mass coordinates from basic spatial coordinates.

        Args:
            Basic spatial coordinates

        Returns:
            Basic mass coordinates
        """

        # Set the mass coordinates at the inner boundary from the core mass
        basic_mass_coordinates = np.zeros_like(basic_coordinates)
        basic_mass_coordinates[:,:] = (
            self.settings.core_density / self._planet_density
            * np.power(basic_coordinates[0,:], 3.0)
        )

        # Get mass coordinates by adding individual cell contributions to the mantle mass
        basic_volumes = (np.power(basic_coordinates[1:,:],3.0)
            - np.power(basic_coordinates[:-1,:],3.0))
        for i in range(1, self.settings.number_of_nodes):
            basic_mass_coordinates[i:,:] += (
                self.staggered_effective_density[i-1,:] * basic_volumes[i-1,:] / self._planet_density
            )

        return np.power(basic_mass_coordinates, 1.0/3.0)

    def get_staggered_spatial_coordinates_from_mass_coordinates(self, staggered_mass_coordinates: npt.NDArray) -> npt.NDArray:
        """Computes the staggered spatial coordinates from staggered mass coordinates.

        Args:
            Staggered mass coordinates

        Returns:
            Staggered spatial coordinates
        """

        # Initialise the staggered spatial coordinate to the inner boundary
        staggered_coordinates = (np.ones_like(staggered_mass_coordinates)
            * np.power(self.settings.inner_radius,3.0))

        # Add first half cell contribution
        staggered_coordinates += (
            self._planet_density
             * (np.power(staggered_mass_coordinates[0,:],3.0)
            - np.power(self.basic.mass_radii[0,:], 3.0))
            / self.staggered_effective_density[0,:]
        )

        # Get spatial coordinates by adding individual cell contributions to the mantle mass
        shell_effective_density = 0.5*(
            self.staggered_effective_density[1:,:] + self.staggered_effective_density[:-1,:])
        shell_mass_volumes = (np.power(staggered_mass_coordinates[1:,:],3.0)
            - np.power(staggered_mass_coordinates[:-1,:],3.0))
        for i in range(1,self.settings.number_of_nodes-1):
            staggered_coordinates[i:,:] += (
                self._planet_density
                * shell_mass_volumes[i-1,:]
                / shell_effective_density[i-1,:]
            )

        return np.power(staggered_coordinates, 1.0/3.0)

    def get_dxidr_basic(self) -> npt.NDArray:
        """Computes dxidr at basic nodes."""

        dxidr = (
            self.eos.basic_density
            / self._planet_density
            * np.power(self.basic.radii,2.0)
            / np.power(self.basic.mass_radii,2.0)
        )

        return dxidr

    def get_constant_spacing(self) -> npt.NDArray:
        """Constant radius spacing across the mantle

        Returns:
            Radii with constant spacing as a column vector
        """
        radii: npt.NDArray = np.linspace(
            self.settings.inner_radius, self.settings.outer_radius, self.settings.number_of_nodes
        )
        radii = np.atleast_2d(radii).T

        return radii

    def _get_d_dr_transform_matrix(self) -> npt.NDArray:
        """Transform matrix for determining d/dr of a staggered quantity on the basic mesh.

        Returns:
            The transform matrix
        """
        transform: npt.NDArray = np.zeros(
            (self.basic.number_of_nodes, self.staggered.number_of_nodes)
        )
        transform[1:-1, :-1] += np.diagflat(-1 / self.staggered.delta_mesh)  # k=0 diagonal
        transform[1:-1:, 1:] += np.diagflat(1 / self.staggered.delta_mesh)  # k=1 diagonal

        # Gradient at boundaries can be extrapolated from the first two closests basic nodes
        # This only affects the estimation of indivual components of heat fluxes when working
        # with flux boundary conditions. Gradient at boundaries are overwritten when using
        # temperature boundary conditions.

        # Extrapolation of gradient at inner radius
        inner_delta_ratio = self.basic.delta_mesh[1].item() / self.basic.delta_mesh[0].item()
        transform[0, 0] = - (inner_delta_ratio + 1) / self.staggered.delta_mesh[0].item()
        transform[0, 1] = (inner_delta_ratio + 1) / self.staggered.delta_mesh[0].item()
        transform[0, 1] += inner_delta_ratio / self.staggered.delta_mesh[1].item()
        transform[0, 2] = - inner_delta_ratio / self.staggered.delta_mesh[1].item()
        # Extrapolation of gradient at outer radius
        outer_delta_ratio: float = self.basic.delta_mesh[-2].item() / self.basic.delta_mesh[-1].item()
        transform[-1, -1] = - (outer_delta_ratio + 1) / self.staggered.delta_mesh[-1].item()
        transform[-1, -2] = (outer_delta_ratio + 1) / self.staggered.delta_mesh[-1].item()
        transform[-1, -2] += outer_delta_ratio / self.staggered.delta_mesh[-2].item()
        transform[-1, -3] = - outer_delta_ratio / self.staggered.delta_mesh[-2].item()

        # Scale the transform matrix by dxi/dr at basic nodes
        for i in range(self.settings.number_of_nodes-1):
            transform[:,i] *= self._dxidr[:, 0]
        logger.debug("_d_dr_transform_matrix = %s", transform)

        return transform

    def d_dr_at_basic_nodes(self, staggered_quantity: npt.NDArray) -> npt.NDArray:
        """Determines d/dr at the basic nodes of a quantity defined at the staggered nodes.

        Args:
            staggered_quantity: A quantity defined at the staggered nodes.

        Returns:
            d/dr at the basic nodes
        """
        d_dr_at_basic_nodes: npt.NDArray = self._d_dr_transform.dot(staggered_quantity)
        logger.debug("d_dr_at_basic_nodes = %s", d_dr_at_basic_nodes)

        return d_dr_at_basic_nodes

    def _get_quantity_transform_matrix(self) -> npt.NDArray:
        """A transform matrix for mapping quantities on the staggered mesh to the basic mesh.

        Uses backward and forward differences at the inner and outer radius, respectively, to
        obtain the quantity values of the basic nodes at the innermost and outermost nodes.
        When using temperature boundary conditions, values at outer boundaries will be overwritten.
        When using flux boundary conditions, values at outer boundaries will be used to provide
        estimate of individual components of heat fluxes though the total heat flux is imposed.

        Returns:
            The transform matrix
        """
        transform: npt.NDArray = np.zeros(
            (self.basic.number_of_nodes, self.staggered.number_of_nodes)
        )
        mesh_ratio: npt.NDArray = self.basic.delta_mesh[:-1] / self.staggered.delta_mesh
        transform[1:-1, :-1] += np.diagflat(1 - 0.5 * mesh_ratio)  # k=0 diagonal
        transform[1:-1:, 1:] += np.diagflat(0.5 * mesh_ratio)  # k=1 diagonal
        # Backward difference at inner radius
        transform[0, :2] = np.array([1 + 0.5 * mesh_ratio[0], -0.5 * mesh_ratio[0]]).flatten()
        # Forward difference at outer radius
        mesh_ratio_outer: npt.NDArray = self.basic.delta_mesh[-1] / self.staggered.delta_mesh[-1]
        transform[-1, -2:] = np.array(
            [-0.5 * mesh_ratio_outer, 1 + 0.5 * mesh_ratio_outer]
        ).flatten()
        logger.debug("_quantity_transform_matrix = %s", transform)

        return transform

    def quantity_at_basic_nodes(self, staggered_quantity: npt.NDArray) -> npt.NDArray:
        """Determines a quantity at the basic nodes that is defined at the staggered nodes.

        Uses backward and forward differences at the inner and outer radius, respectively, to
        obtain the quantity values of the basic nodes at the innermost and outermost nodes.
        When using temperature boundary conditions, values at outer boundaries will be overwritten.
        When using flux boundary conditions, values at outer boundaries will be used to provide
        estimate of individual components of heat fluxes though the total heat flux is imposed.

        Args:
            staggered_quantity: A quantity defined at the staggered nodes

        Returns:
            The quantity at the basic nodes
        """
        quantity_at_basic_nodes: npt.NDArray = self._quantity_transform.dot(staggered_quantity)
        logger.debug("quantity_at_basic_nodes = %s", quantity_at_basic_nodes)

        return quantity_at_basic_nodes

    def quantity_at_staggered_nodes(self, basic_quantity: npt.NDArray) -> npt.NDArray:
        """Determines a quantity at the staggered nodes that is defined at the basic nodes.

        Staggered nodes are always located at cell centers, whatever the mesh.

        Args:
            basic_quantity: A quantity defined at the basic nodes

        Returns:
            The quantity at the staggered nodes
        """
        quantity_at_staggered_nodes: npt.NDArray = 0.5 * (
            basic_quantity[:-1, ...] + basic_quantity[1:, ...])
        logger.debug("quantity_at_staggered_nodes = %s", quantity_at_staggered_nodes)

        return quantity_at_staggered_nodes

    def volume_average(self, staggered_quantity: npt.NDArray) -> float:
        return float(np.dot(staggered_quantity.T, self.basic.volume)) / self.basic.total_volume


class EOS(ABC):
    """Generic EOS class"""

    @abstractmethod
    def staggered_effective_density(self) -> npt.NDArray: ...

    @abstractmethod
    def basic_density(self) -> npt.NDArray: ...

    @abstractmethod
    def basic_pressure(self) -> npt.NDArray: ...

    @abstractmethod
    def staggered_pressure(self) -> npt.NDArray: ...

    @abstractmethod
    def set_staggered_pressure(self, staggered_radii: npt.NDArray,) -> None: ...

class AdamsWilliamsonEOS(EOS):
    r"""Adams-Williamson equation of state (EOS).

    EOS due to adiabatic self-compression from the definition of the adiabatic bulk modulus:

    .. math::

        \left( \frac{{d\rho}}{{dP}} \right)_S = \frac{{\rho}}{{K_S}}

    where :math:`\rho` is density, :math:`K_S` the adiabatic bulk modulus, and :math:`S` is
    entropy.
    """

    def __init__(
        self,
        settings: _MeshParameters,
        basic_radii: npt.NDArray,
    ):
        self._settings: _MeshParameters = settings
        self._basic_radii: npt.NDArray = basic_radii
        self._outer_boundary = np.max(basic_radii)
        self._inner_boundary = np.min(basic_radii)
        self._surface_density: float = self._settings.surface_density
        self._gravitational_acceleration: float = self._settings.gravitational_acceleration
        self._adiabatic_bulk_modulus: float = self._settings.adiabatic_bulk_modulus
        self._basic_pressure = self.get_pressure_from_radii(basic_radii)
        self._basic_density = self.get_density_from_radii(basic_radii)
        self._staggered_effective_density = self.get_effective_density(basic_radii)

    @property
    def basic_pressure(self) -> npt.NDArray:
        """Pressure at basic nodes"""
        return self._basic_pressure

    @property
    def staggered_pressure(self) -> npt.NDArray:
        """Pressure at staggered nodes"""
        return self._staggered_pressure

    @property
    def basic_density(self) -> npt.NDArray:
        """Density at basic nodes"""
        return self._basic_density

    @property
    def staggered_effective_density(self) -> npt.NDArray:
        """Effective density at staggered nodes"""
        return self._staggered_effective_density

    def set_staggered_pressure(self, staggered_radii: npt.NDArray,) -> None:
        """Set staggered pressure based on staggered radii."""
        self._staggered_pressure = self.get_pressure_from_radii(staggered_radii)

    def get_effective_density(self, radii) -> npt.NDArray:
        r"""
        Computes effective density using density rho(r) integration
        over a spherical shell bounded by radii

        Args:
            radii: Radii array

        Returns:
            Effective Density array
        """

        mass_shell = self.get_mass_within_shell(radii)
        volume_shell = 4 / 3 * np.pi  \
            * (np.power(radii[1:],3.0) - np.power(radii[:-1],3.0))

        return mass_shell/volume_shell

    def get_density(self, pressure: FloatOrArray) -> npt.NDArray:
        r"""Computes density from pressure:

        .. math::

            \rho(P) = \rho_s \exp(P/K_S)

        where :math:`\rho` is density, :math:`P` is pressure, :math:`\rho_s` is surface density,
        and :math:`K_S` is adiabatic bulk modulus.

        Args:
            pressure: Pressure

        Returns:
            Density
        """
        density: npt.NDArray = self._surface_density * np.exp(
            pressure / self._adiabatic_bulk_modulus
        )

        return density

    def get_density_from_radii(self, radii: FloatOrArray) -> FloatOrArray:
        r"""Computes density from radii:

        .. math::

            \rho(r) = \frac{\rho_s K_S}{K_S + \rho_s g (r-r_s)}

        where :math:`\rho` is density, :math:`r` is radius, :math:`\rho_s` is surface density,
        :math:`K_S` is adiabatic bulk modulus, and :math:`r_s` is surface radius.

        Args:
            radii: Radii

        Returns
            Density
        """
        density: FloatOrArray = (self._surface_density * self._adiabatic_bulk_modulus) / (
            self._adiabatic_bulk_modulus
            + self._surface_density
            * self._gravitational_acceleration
            * (radii - self._outer_boundary)
        )

        return density

    def get_mass_element(self, radii: FloatOrArray) -> npt.NDArray:
        r"""Computes the mass element:

        .. math::

            \frac{\delta m}{\delta r} = 4 \pi r^2 \rho

        where :math:`\delta m` is the mass element, :math:`r` is radius, and :math:`\rho` is
        density.

        Args:
            radii: Radii

        Returns:
            The mass element at radii
        """
        mass_element: npt.NDArray = (
            4 * np.pi * np.square(radii) * self.get_density_from_radii(radii)
        )

        return mass_element

    def get_mass_within_radii(self, radii: FloatOrArray) -> npt.NDArray:
        r"""Computes mass within radii:

        .. math::

            m(r) = \int 4 \pi r^2 \rho dr

        where :math:`m` is mass, :math:`r` is radius, and :math:`\rho` is density.

        The integral was evaluated using WolframAlpha.

        Args:
            radii: Radii

        Returns:
            Mass within radii
        """
        a: float = self._surface_density
        b: float = self._adiabatic_bulk_modulus
        c: float = self._gravitational_acceleration
        d: float = self._outer_boundary
        beta: float = b / (a * c) - d

        def mass_integral(radii_: FloatOrArray) -> npt.NDArray:
            """Mass within radii including arbitrary constant of integration.

            Args:
                radii_: Radii

            Returns:
                Mass within radii
            """

            mass: npt.NDArray = (
                4
                * np.pi
                * b
                / c
                * (
                    -1.5 * beta * beta
                    - beta * radii_
                    + 0.5 * radii_ * radii_
                    + beta * beta * np.log(abs(beta + radii_))
                )
            )
            # + constant

            return mass

        mass: npt.NDArray = mass_integral(radii) - mass_integral(self._inner_boundary)

        return mass

    def get_mass_within_shell(self, radii: npt.NDArray) -> npt.NDArray:
        """Computes the mass within spherical shells bounded by radii.

        Args:
            radii: Radii

        Returns:
            Mass within the bounded spherical shells
        """
        mass: npt.NDArray = self.get_mass_within_radii(radii[1:]) - self.get_mass_within_radii(
            radii[:-1]
        )

        return mass

    def get_pressure_from_radii(self, radii: FloatOrArray) -> npt.NDArray:
        r"""Computes pressure from radii:

        .. math::

            P(r) = -K_S \ln \left( 1 + \frac{\rho_s g (r-r_s)}{K_S} \right)

        where :math:`r` is radius, :math:`K_S` is adiabatic bulk modulus, :math:`P` is pressure,
        :math:`\rho_s` is surface density, :math:`g` is gravitational acceleration, and
        :math:`r_s` is surface radius.

        Args:
            radii: Radii

        Returns:
            Pressure
        """
        pressure: npt.NDArray = -self._adiabatic_bulk_modulus * np.log(
            (
                self._adiabatic_bulk_modulus
                + self._surface_density
                * self._gravitational_acceleration
                * (radii - self._outer_boundary)
            )
            / self._adiabatic_bulk_modulus
        )

        return pressure

    def get_pressure_gradient(self, pressure: FloatOrArray) -> npt.NDArray:
        r"""Computes the pressure gradient:

        .. math::

            \frac{dP}{dr} = -g \rho

        where :math:`\rho` is density, :math:`P` is pressure, and  :math:`g` is gravitational
        acceleration.

        Args:
            pressure: Pressure

        Returns:
            Pressure gradient
        """
        dPdr: npt.NDArray = -self._gravitational_acceleration * self.get_density(pressure)

        return dPdr

    def get_radii_from_pressure(self, pressure: FloatOrArray) -> npt.NDArray:
        r"""Computes radii from pressure:

        .. math::

            P(r) = \int \frac{dP}{dr} dr = \int -g \rho_s \exp(P/K_S) dr

        And apply the boundary condition :math:`P=0` at :math:`r=r_s` to get:

        .. math::

            r(P) = \frac{K_s \left( \exp(-P/K_S)-1 \right)}{\rho_s g} + r_s

        where :math:`r` is radius, :math:`K_S` is adiabatic bulk modulus, :math:`P` is pressure,
        :math:`\rho_s` is surface density, :math:`g` is gravitational acceleration, and
        :math:`r_s` is surface radius.

        Args:
            pressure: Pressure

        Returns:
            Radii
        """
        radii: npt.NDArray = (
            self._adiabatic_bulk_modulus
            * (np.exp(-pressure / self._adiabatic_bulk_modulus) - 1)
            / (self._surface_density * self._gravitational_acceleration)
            + self._outer_boundary
        )

        return radii

class UserDefinedEOS(EOS):
    r"""User defined equation of state (EOS).

    Pressure field and effective density field on staggered nodes provided by the user.
    """

    def __init__(
        self,
        settings: _MeshParameters,
        basic_radii: npt.NDArray,
    ):
        self._interp_pressure = PchipInterpolator(settings.eos_radius, settings.eos_pressure)
        self._interp_density = PchipInterpolator(settings.eos_radius, settings.eos_density)
        self._basic_pressure = self._interp_pressure(basic_radii).reshape(-1,1)
        basic_effective_density = self._interp_density(basic_radii).reshape(-1,1)
        self._staggered_effective_density = 0.5*(
            basic_effective_density[:-1, :] + basic_effective_density[1:, :])
        # Assumes density and effective density are the same at basic nodes
        self._basic_density = basic_effective_density

    @property
    def basic_pressure(self) -> npt.NDArray:
        """Pressure at basic nodes"""
        return self._basic_pressure

    @property
    def staggered_pressure(self) -> npt.NDArray:
        """Pressure at staggered nodes"""
        return self._staggered_pressure

    @property
    def basic_density(self) -> npt.NDArray:
        """Effective density at basic nodes"""
        return self._basic_density

    @property
    def staggered_effective_density(self) -> npt.NDArray:
        """Effective density at staggered nodes"""
        return self._staggered_effective_density

    def set_staggered_pressure(self, staggered_radii: npt.NDArray,) -> None:
        """Set staggered pressure based on staggered radii."""
        self._staggered_pressure = self._interp_pressure(staggered_radii).reshape(-1,1)
